#!/usr/bin/env python
# Copyright 2016 Henning Glawe
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""FPLO uses C-inspired input files. They are not quite C, so no conventional
C parser library can be used.
Among the more complex features are nested structs, as well
as arrays-of-struct.

This module is implemented as follows:
1) tokenizer for the used C subset/dialect
2) transformation of tokenized output to concrete syntax tree
3) transformation of concrete syntax tree to abstract syntax tree (AST)
TODO:
4) transform AST to metaInfo backend calls
"""
import re
import sys
import os
import logging
import json
from . import FploCommon as FploC
from nomadcore.match_highlighter import ANSI
from nomadcore.caching_backend import CachingLevel, ActiveBackend

LOGGER = logging.getLogger(__name__)


class TokenMatchError(Exception):
    pass


class token(object):
    highlight_start = ''
    highlight_end = ANSI.RESET
    regex = None
    cRE_end_newline = re.compile(r'(.*?)(\n*)$')

    def __init__(self, line, pos_in_line):
        """token constructor takes re.match object as arg"""
        match = self.regex.match(line, pos_in_line)
        if match is None:
            raise TokenMatchError
        self._match = match
        self.value = self.match2value()
        self.source_line = line

    def highlighted(self):
        """return ANSI-highlighted token"""
        m = self.cRE_end_newline.match(self._match.group(0))
        return self.highlight_start + m.group(1) + self.highlight_end + m.group(2)

    def match_end(self):
        return self._match.end()

    def match2value(self):
        return None

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return "%10s %s" % (self.__class__.__name__, repr(self.value))


class token_literal(token):
    regex = re.compile(
        r'\s*' + r'(?:' + r'|'.join([
            # alternates for literals
            # RE_f,
            r'"(?P<str_d>[^"\\]*(?:\\\\|\\"|[^"]*)*)"',
            r"'(?P<str_s>[^'\\]*(?:\\\\|\\'|[^']*)*)'",
            r'(?P<float>' + (
                r'[+-]?' + # optional sign
                r'\d+(?=[\.eE])' + # positive lookahead: either decimal point or exponential part must follow
                r'(?:\.\d*)?' + #cover decimals if present
                r'(?:[eE][+-]\d+)?' # exponential part if present
            r')'),
            r'0x(?P<hex_int>[0-9a-fA-F]+)',
            r'0(?P<octal_int>[0-7]+)',
            r'(?P<decimal_int>[+-]?\d+)', # integer with optional sign
            r'(?P<logical>[tf])(?=\W)',
        ]) + r')'
    )

    def match2value(self):
        match = self._match
        if match.group('str_d') is not None:
            return match.group('str_d')
        if match.group('str_s') is not None:
            return match.group('str_s')
        if match.group('float') is not None:
            return float(match.group('float'))
        if match.group('hex_int') is not None:
            return int(match.group('hex_int'), base=16)
        if match.group('octal_int') is not None:
            return int(match.group('octal_int'), base=8)
        if match.group('decimal_int') is not None:
            return int(match.group('decimal_int'))
        if match.group('logical') is not None:
            if match.group('logical') == 't':
                return True
            else:
                return False
        raise RuntimeError('no idea what to do with literal "%s"' % (match.group(0)))


class token_datatype(token):
    regex = re.compile(r'\s*([a-zA-Z_][a-zA-Z0-9_]*)')
    subtype_list = []
    subtype_dict = {}

    def match2value(self):
        value_index = self.subtype_dict.get(self._match.group(1), None)
        if value_index is None:
            raise TokenMatchError
        self.value_index = value_index
        value = self.subtype_list[value_index]
        return value

token_datatype.subtype_list = [
        'char',
        'int',
        'real',
        'logical',
        'flag',
    ]

token_datatype.subtype_dict = { token_datatype.subtype_list[i]: i for i in range(len(token_datatype.subtype_list)) }


class token_keyword(token):
    regex = re.compile(r'\s*([a-zA-Z_][a-zA-Z0-9_]*)')
    subtype_list = []
    subtype_dict = {}

    def match2value(self):
        value_index = self.subtype_dict.get(self._match.group(1), None)
        if value_index is None:
            raise TokenMatchError
        self.value_index = value_index
        value = self.subtype_list[value_index]
        return value

token_keyword.subtype_list = [
        'section',
        'struct',
    ]

token_keyword.subtype_dict = { token_keyword.subtype_list[i]: i for i in range(len(token_keyword.subtype_list)) }


class token_identifier(token):
    regex = re.compile(r'\s*([a-zA-Z_][a-zA-Z0-9_]*)')

    def match2value(self):
        return self._match.group(1)


class token_subscript_begin(token):
    regex = re.compile(r'\[')


class token_subscript_end(token):
    regex = re.compile(r'\]')


class token_operator(token):
    regex = re.compile(r'\s*(\+=|\-=|=|,|-|\+|/|\*)')

    def match2value(self):
        return self._match.group(1)


class token_block_begin(token):
    regex = re.compile(r'\s*\{')


class token_block_end(token):
    regex = re.compile(r'\s*\}')


class token_statement_end(token):
    regex = re.compile(r'\s*;')


class token_line_comment(token):
    regex = re.compile(r'\s*(?:(//|#)|(/\*))(?P<comment>.*)')

    def match2value(self):
        return self._match.group('comment')

class token_trailing_whitespace(token):
    regex = re.compile(r'\s+$')

class token_bad_input(token):
    regex = re.compile('(.+)$')

    def match2value(self):
        return self._match.group(1)

class token_flag_value(token):
    regex = re.compile(r'\(([+-])\)')

    def match2value(self):
        if self._match.group(1) == '+':
            return True
        else:
            return False


token_literal.highlight_start = ANSI.FG_MAGENTA
token_datatype.highlight_start = ANSI.FG_YELLOW
token_keyword.highlight_start = ANSI.FG_BRIGHT_YELLOW
token_identifier.highlight_start = ANSI.FG_CYAN
token_subscript_begin.highlight_start = ANSI.FG_BRIGHT_GREEN
token_subscript_end.highlight_start = ANSI.FG_BRIGHT_GREEN
token_operator.highlight_start = ANSI.BEGIN_INVERT + ANSI.FG_YELLOW
token_block_begin.highlight_start = ANSI.FG_BRIGHT_CYAN
token_block_end.highlight_start = ANSI.FG_BRIGHT_CYAN
token_statement_end.highlight_start = ANSI.FG_BRIGHT_YELLOW
token_line_comment.highlight_start = ANSI.FG_BLUE
token_trailing_whitespace.highlight_start = ANSI.BG_BLUE
token_bad_input.highlight_start = ANSI.BEGIN_INVERT + ANSI.FG_BRIGHT_RED
token_flag_value.highlight_start = ANSI.FG_MAGENTA


class AST_node(dict):
    """base class for abstract syntax tree nodes"""
    def __init__(self, name=None):
        self.name = name
        self.child = []

    def indented_str(self, indent=''):
        result = ANSI.BG_YELLOW + indent + ANSI.RESET + ('%-20s' % (self.__class__.__name__))
        if self.name is not None:
            result = result + ' ' + self.name
        result = result + '\n'
        child_indent = indent + '  '
        for child in self.child:
            if child is not None:
                result = result + child.indented_str(child_indent)
        return result

    def append(self, newchild):
        self.child.append(newchild)

    def __len__(self):
        return len(self.child)

    def declaration_nomadmetainfo(self, output_file, namespace):
        if self.name is not None:
            child_namespace = namespace + '.' + self.name
        else:
            child_namespace = namespace
        for child in self.child:
            if isinstance(child, AST_node):
                child.declaration_nomadmetainfo(output_file, child_namespace)

    def data_nomadmetainfo(self, backend, namespace):
        if self.name is not None:
            child_namespace = namespace + '.' + self.name
        else:
            child_namespace = namespace
        for child in self.child:
            if isinstance(child, AST_node):
                child.data_nomadmetainfo(backend, child_namespace)



class AST_block(AST_node):
    """generic block (sequence of statements) in AST"""
    def append_block(self, src_block):
        for src_child in src_block.child:
            self.append(src_child)


class AST_root(AST_block):
    def declaration_nomadmetainfo(self, output_file, namespace):
        output_file.write((
            '{\n' +
            '  "type": "nomad_meta_info_1_0",\n' +
            '  "description": "FPLO input metainfo, autogenerated",\n' +
            '  "dependencies": [{\n' +
            '    "relativePath": "public.nomadmetainfo.json"\n' +
            '  }],\n' +
            '  "metaInfos": [\n' +
            '    {\n' +
            '      "description": "FPLO input metainfo, autogenerated",\n' +
            '      "name": "%s",\n' +
            '      "kindStr": "type_section",\n' +
            '      "superNames": [ "section_method" ]\n' +
            '    }') % (namespace))
        AST_block.declaration_nomadmetainfo(self, output_file, namespace)
        output_file.write(
            '\n' +
            '  ]\n' +
            '}\n'
        )

    def data_nomadmetainfo(self, backend, namespace):
        gIndex = backend.openSection(namespace)
        AST_node.data_nomadmetainfo(self, backend, namespace)
        backend.closeSection(namespace, gIndex)


class AST_section(AST_block):
    """section block (sequence of statements) in AST"""
    def declaration_nomadmetainfo(self, output_file, namespace):
        thisname = namespace + '.' + self.name
        output_file.write((
            ', {\n' +
            '      "description": "FPLO input section %s",\n' +
            '      "name": "%s",\n' +
            '      "kindStr": "type_section",\n' +
            '      "superNames": [ "%s" ]\n' +
            '    }') % (thisname, thisname, namespace))
        AST_block.declaration_nomadmetainfo(self, output_file, namespace)


class AST_datatype(AST_node):
    def nomad_kindStr(self):
        return None

    def nomad_dtypeStr(self):
        return None


class AST_datatype_primitive(AST_datatype):
    dtype2nomad = {
        'char': 'C',
        'int': 'i',
        'real': 'f',
        'logical': 'b',
    }

    def nomad_dtypeStr(self):
        return self.dtype2nomad[self.name]


class AST_datatype_struct(AST_datatype):
    def append_block(self, src_block):
        for src_child in src_block.child:
            self.append(src_child)

    def nomad_kindStr(self):
        return 'type_section'


class AST_datatype_flag(AST_datatype_struct):
    def nomad_kindStr(self):
        return 'type_section'


class AST_declaration(AST_node):
    """variable declaration in abstract syntax tree"""
    # children:
    #   0 - shape
    #   1 - datatype
    def __init__(self, name, datatype, shape=None):
        AST_node.__init__(self, name)
        self.child.append(shape)
        self.child.append(datatype)

    def set_shape(self, shape):
        if self.child[0] is not None:
            raise RuntimeError('already has shape: %s', self.name)
        self.child[0] = shape

    def declaration_nomadmetainfo(self, output_file, namespace):
        thisname = namespace + '.' + self.name
        kindStr = self.child[1].nomad_kindStr()
        dtypeStr = self.child[1].nomad_dtypeStr()
        output_file.write((
            ', {\n' +
            '      "description": "FPLO input %s",\n' +
            '      "name": "%s",\n' +
            '      "superNames": [ "%s" ],\n'
            ) % (thisname, thisname, namespace))
        if self.child[0] is not None:
            output_file.write('      "repeats": true,\n')
        if kindStr is not None:
            output_file.write('      "kindStr": "%s"\n' % (kindStr))
        elif dtypeStr is not None:
            output_file.write('      "dtypeStr": "%s"\n' % (dtypeStr))
        else:
            raise RuntimeError(
                "neither kindStr nor dtypeStr are defined for %s" % (
                    thisname
               )
            )
        output_file.write('    }')
        AST_node.declaration_nomadmetainfo(self, output_file, namespace)


class AST_shape(AST_node):
    # children are ints without indented_str method
    def indented_str(self, indent=''):
        result = ANSI.BG_YELLOW + indent + ANSI.RESET + ('%-20s [' % (self.__class__.__name__))
        result = result + ', '.join(map(str, self.child))
        result = result + ']\n'
        return result


class AST_value(AST_node):
    pass


class AST_value_primitive(AST_value):
    # there is one child, a python literal
    def indented_str(self, indent=''):
        result = ANSI.BG_YELLOW + indent + ANSI.RESET + (
            '%-20s %s\n' % (self.__class__.__name__, str(self.child[0])))
        return result


class AST_value_list(AST_value):
    def indented_str(self, indent=''):
        result = ANSI.BG_YELLOW + indent + ANSI.RESET + (
            '%-20s %s\n' % (self.__class__.__name__, repr(self.child)))
        return result


class AST_assignment(AST_node):
    # children:
    # 0: target (type AST_declaration)
    # 1: value (type AST_value)
    pass


class concrete_node(object):
    def __init__(self, parent, source_line):
        self.items = []
        self.source_line = source_line
        # backref
        self.parent = parent

    def append(self, item):
        self.items.append(item)

    def indented_dump(self, indent):
        if len(self.items) < 1:
            return ''
        result = indent + self.__class__.__name__ + ":\n"
        for item in self.items:
            if isinstance(item, concrete_node):
                result = result + item.indented_dump(indent + '  ')
            else:
                result = result + indent + '  ' + str(item) + '\n'
        return result

    def to_AST(self):
        raise Exception("unimplemented to_AST in %s" % (self.__class__.__name__))


class concrete_statement(concrete_node):
    def to_AST(self):
        if len(self.items) < 1:
            return None
        result = None
        pos_in_statement = 0
        # check declarations
        if isinstance(self.items[0], token_keyword):
            if self.items[0].value == 'section':
                result = AST_section(self.items[1].value) # name of section
                result.append_block(self.items[2].to_AST()) # section-block
                pos_in_statement = 3
            if self.items[0].value == 'struct':
                struct = AST_datatype_struct()
                struct.append_block(self.items[1].to_AST())
                result = AST_declaration(self.items[2].value, struct)
                pos_in_statement = 3
        elif isinstance(self.items[0], token_datatype) and self.items[0].value == 'flag':
            # special case for non-C-primtype 'flag'
            #   we will map this to struct of logicals, but need to evaluate
            #   RHS to get the names. messy.
            flag = AST_datatype_flag()
            result = AST_declaration(self.items[1].value, flag)
            if isinstance(self.items[2], concrete_subscript):
                # skip array shape
                pos_in_statement = 3
            else:
                pos_in_statement = 2
        elif isinstance(self.items[0], token_datatype):
            primtype = AST_datatype_primitive(self.items[0].value)
            if primtype.name == 'char' and isinstance(self.items[1], concrete_subscript):
                # ignore char length for now
                #   not correct in C, but all declared chars in FPLO input
                #   are char arrays
                declaration_name = self.items[2].value
                pos_in_statement = 3
            else:
                declaration_name = self.items[1].value
                pos_in_statement = 2
            result = AST_declaration(declaration_name, primtype)
        if (
                (len(self.items) > pos_in_statement) and
                isinstance(self.items[pos_in_statement], concrete_subscript)
            ):
            # subscript in LHS declares shape
            if not isinstance(result, AST_declaration):
                raise RuntimeError('encountered subscript on non-declaration')
            result.set_shape(self.items[pos_in_statement].to_AST_shape())
            pos_in_statement = pos_in_statement + 1
        if len(self.items) <= pos_in_statement:
            # we are done, nothing more in statement
            return result
        if not (isinstance(self.items[pos_in_statement], token_operator) and
                self.items[pos_in_statement].value == '='):
            raise RuntimeError('unexpected item following declaration: "%s", line: "%s"' % (
                repr(self.items[pos_in_statement]), self.items[pos_in_statement].source_line))
        # we have an assignment
        new_assignment = AST_assignment()
        new_assignment.append(result)
        pos_in_statement = pos_in_statement + 1
        if len(self.items) <= pos_in_statement:
            raise RuntimeError('missing values in assignment')
        if len(self.items) > pos_in_statement + 1:
            raise RuntimeError('too many values in assignment')
        concrete_values = self.items[pos_in_statement]
        if isinstance(concrete_values, token_literal):
            new_value = AST_value_primitive()
            new_value.append(concrete_values.value)
            new_assignment.append(new_value)
            return new_assignment
        if isinstance(concrete_values, concrete_block):
            new_value = AST_value_list()
            if isinstance(result.child[1], AST_datatype_flag):
                # special case for 'flag' datatype, need to evaluate RHS for names
                (flag_names, flag_values) = concrete_values.flag_names_values()
                for flag_name in flag_names:
                    result.child[1].append(AST_declaration(flag_name, AST_datatype_primitive('logical')))
                for flag_value in flag_values:
                    new_value.append(flag_value)
            else:
                new_value.child = concrete_values.python_value()
            new_assignment.append(new_value)
        return new_assignment

    def python_value(self):
        def eval_accumulated(items):
            if len(items) == 1:
                return items[0]
            elif len(items) == 3 and items[1] == '/':
                return items[0] / items[2]
            else:
                LOGGER.error('concrete_statement.python_value:accum: %s', str(items))
                sys.stderr.write(self.indented_dump(''))

        result = []
        accum = []
        for item in self.items:
            if isinstance(item, token_literal):
                accum.append(item.value)
            elif isinstance(item, token_operator) and item.value == ',':
                if len(accum) > 0:
                    result.append(eval_accumulated(accum))
                accum = []
            elif isinstance(item, token_operator) and item.value == '/':
                # FPLO input contains fraction constants
                accum.append('/')
            elif isinstance(item, concrete_block):
                result.append(item.python_value())
            else:
                LOGGER.error('concrete_statement.python_value:item: %s', repr(item))
        if len(accum) > 0:
            result.append(eval_accumulated(accum))
        return result

    def flag_names_values(self):
        result_names = []
        result_values = []
        accum = []
        for item in self.items:
            if isinstance(item, (token_identifier, token_flag_value)):
                accum.append(item.value)
            elif isinstance(item, token_operator) and item.value == ',':
                if len(accum)!=2:
                    raise RuntimeError('flag_names_values encountered non-pair: ', str(accum))
                if accum[0] != 'NOT_USED':
                    result_names.append(accum[0])
                    result_values.append(accum[1])
                accum=[]
            else:
                raise RuntimeError('flag_names_values encountered unhandled item: ' + repr(item))
        if len(accum) > 0:
            if len(accum)!=2:
                raise RuntimeError('flag_names_values encountered non-pair: ', str(accum))
            if accum[0] != 'NOT_USED':
                result_names.append(accum[0])
                result_values.append(accum[1])
        return (result_names, result_values)

class concrete_block(concrete_node):
    def to_AST(self):
        if len(self.items) < 1:
            return None
        result = AST_block()
        for item in self.items:
            item_AST = item.to_AST()
            if item_AST is not None:
                result.append(item_AST)
        if len(result) is not None:
            return result
        return None

    def python_value(self):
        if len(self.items) != 1:
            raise RuntimeError('python_value for block containing !=1 statement')
        return self.items[0].python_value()

    def flag_names_values(self):
        if len(self.items) != 1:
            raise RuntimeError('flag_names for block containing !=1 statement')
        return self.items[0].flag_names_values()


class concrete_root(concrete_block):
    def to_AST(self):
        if len(self.items) < 1:
            return None
        result = AST_root()
        for item in self.items:
            item_AST = item.to_AST()
            if item_AST is not None:
                result.append(item_AST)
        if len(result) is not None:
            return result
        return None


class concrete_subscript(concrete_statement):
    def __str__(self):
        result = (
            '[' +
            ', '.join(map(lambda i: "'" + str(i) + "'", self.items)) +
            ']'
        )
        return result

    def __repr__(self):
        result = (
            '[' +
            ', '.join(map(lambda i: "'" + repr(i) + "'", self.items)) +
            ']'
        )
        return "%10s %s" % (self.__class__.__name__, result)

    def to_AST_shape(self):
        result = AST_shape()
        for item in self.items:
            if isinstance(item, token_literal) and isinstance(item.value, int):
                result.append(item.value)
            elif isinstance(item, token_operator) and item.value == '*':
                # denote variable-length dimension by -1
                result.append(int(-1))
            elif isinstance(item, token_identifier):
                # TODO: check if length from identifier needs to be respected
                # for now treat as variable-length
                result.append(int(-1))
            else:
                raise Exception("unknown purpose of item in shape: %s" % (repr(item)))
        return result


class FploInputParser(object):
    """Parser for C-like FPLO input
    """
    def __init__(self, file_path, annotateFile = None,
                 annotated_line_callback = None):
        self.input_tree = {}
        self.file_path = file_path
        self.state = self.state_root
        self.__annotateFile = annotateFile
        self.bad_input = False
        # start with root block, and add empty statement to append to
        self.concrete_statements = concrete_root(None, '')
        self.concrete_statements.append(concrete_statement(self.concrete_statements, ''))
        self.current_concrete_statement = self.concrete_statements.items[-1]
        self.annotated_line = ''
        self.annotated_line_callback = annotated_line_callback

    def parse(self):
        """open file and parse line-by-line"""
        with open(self.file_path, "r") as fIn:
            # process line-by-line
            for line in fIn:
                self.parse_line(line)
        # check if there was input flagged as 'bad'/'syntactically incorrect'
        if self.bad_input:
            # call bad-input hook
            self.onBad_input()
        # call end-of-file hook
        self.onEnd_of_file()

    def parse_line(self, line):
        """parse one line, delegating to the parser state handlers"""
        pos_in_line = 0
        while pos_in_line<len(line):
            new_pos_in_line = self.state(line, pos_in_line)
            # check if anything was parsed, otherwise cancel that line
            if new_pos_in_line is None:
                break
            else:
                pos_in_line = new_pos_in_line

    def _annotate(self, what):
        """write string to annotateFile if present"""
        self.annotated_line = self.annotated_line + what
        if self.annotated_line[-1] == '\n':
            if self.__annotateFile:
                self.__annotateFile.write(self.annotated_line)
            if self.annotated_line_callback is not None:
                self.annotated_line_callback(self.annotated_line)
            self.annotated_line = ''

    def state_root(self, line, pos_in_line):
        """state: no open section, i.e. at the root of the namelist"""
        this_token = None
        for try_token in [token_literal, token_flag_value, token_datatype,
                          token_keyword,
                          token_identifier, token_subscript_begin,
                          token_subscript_end, token_operator,
                          token_block_begin, token_block_end,
                          token_statement_end, token_line_comment,
                          token_trailing_whitespace,
                          token_bad_input,
                          ]:
            try:
                this_token = try_token(line, pos_in_line)
            except TokenMatchError:
                pass
            if this_token is not None:
                break
        if this_token is None:
            LOGGER.error("cannot match any token type to '%s'" % (
                line[pos_in_line:]))
            return None
        self._annotate(this_token.highlighted())
        if isinstance(this_token, token_block_begin):
            newblock = concrete_block(self.current_concrete_statement, line)
            newblock.append(concrete_statement(newblock, line))
            self.current_concrete_statement.append(newblock)
            self.current_concrete_statement = newblock.items[0]
        elif isinstance(this_token, token_block_end):
            self.current_concrete_statement = self.current_concrete_statement.parent.parent
        elif isinstance(this_token, token_subscript_begin):
            newsubscript = concrete_subscript(self.current_concrete_statement, line)
            self.current_concrete_statement.append(newsubscript)
            self.current_concrete_statement = newsubscript
        elif isinstance(this_token, token_subscript_end):
            self.current_concrete_statement = self.current_concrete_statement.parent
        elif isinstance(this_token, token_statement_end):
            self.current_concrete_statement.parent.append(concrete_statement(self.current_concrete_statement.parent, line))
            self.current_concrete_statement = self.current_concrete_statement.parent.items[-1]
        elif isinstance(this_token, token_bad_input):
            self.bad_input = True
        elif isinstance(this_token, (token_line_comment, token_trailing_whitespace)):
            # skip comments and trailing whitespace
            pass
        elif isinstance(this_token, (
                    token_literal, token_flag_value, token_operator,
                    token_datatype, token_keyword, token_identifier
                )):
            self.current_concrete_statement.append(this_token)
        else:
            raise Exception("Unhandled token type " + this_token.__class__.__name__)
        return this_token.match_end()

    def onBad_input(self):
        """hook: called at the end of parsing if there was any bad input"""
        pass

    def onEnd_of_file(self):
        """hook: called at the end of parsing"""
        self.AST = self.concrete_statements.to_AST()

if __name__ == "__main__":
    import argparse
    ARGPARSER = argparse.ArgumentParser(description='NOMAD parser for FPLO input')
    ARGPARSER.add_argument('--annotate', action='store_true', default=False,
                           help='write annotated/tokenized input files to stderr')
    ARGPARSER.add_argument('--dump_concrete', action='store_true', default=False,
                           help='write annotated/tokenized input files to stderr')
    ARGPARSER.add_argument('--dump_ast', action='store_true', default=False,
                           help='write annotated/tokenized input files to stderr')
    ARGPARSER.add_argument('--autogenerate_metainfo', action='store_true', default=False,
                           help='write nomadmetainfo to stdout')
    ARGPARSER.add_argument('fplo_input', type=str, nargs='+', help='FPLO input files')
    ARGS = ARGPARSER.parse_args()

    cachingLevelForMetaName = {}
    for name in FploC.META_INFO.infoKinds:
        # set all temporaries to caching-only
        if name.startswith('x_fplo_t_'):
            cachingLevelForMetaName[name] = CachingLevel.Cache

    if ARGS.annotate:
        ANNOTATEFILE = sys.stderr
    else:
        ANNOTATEFILE = None
    for fplo_in in ARGS.fplo_input:
        parser = FploInputParser(fplo_in, annotateFile=ANNOTATEFILE)
        parser.parse()
        sys.stdout.flush()
        sys.stderr.flush()
        if ARGS.dump_concrete:
            sys.stderr.write('concrete syntax tree:\n')
            sys.stderr.write(parser.concrete_statements.indented_dump('  '))
            sys.stdout.flush()
            sys.stderr.flush()
        if ARGS.dump_ast:
            sys.stderr.write('abstract syntax tree:\n')
            sys.stderr.write(parser.AST.indented_str('  '))
            sys.stdout.flush()
            sys.stderr.flush()
        if ARGS.autogenerate_metainfo:
            parser.AST.declaration_nomadmetainfo(sys.stdout, 'x_fplo_in')
        else:
            outF = sys.stdout
            jsonBackend = JsonParseEventsWriterBackend(FploC.META_INFO, outF)
            sys.stderr.write('WTF2:' + str(FploC.META_INFO) + '\n')
            backend = ActiveBackend.activeBackend(
                metaInfoEnv = FploC.META_INFO,
                cachingLevelForMetaName = cachingLevelForMetaName,
                defaultDataCachingLevel = CachingLevel.ForwardAndCache,
                defaultSectionCachingLevel = CachingLevel.Forward,
                superBackend = jsonBackend)
            outF.write("[")
            backend.startedParsingSession('file://' + fplo_in, FploC.PARSER_INFO_DEFAULT)
            run_gIndex = backend.openSection('section_run')
            method_gIndex = backend.openSection('section_method')
            parser.AST.data_nomadmetainfo(backend, 'x_fplo_in')
            backend.closeSection('section_method', method_gIndex)
            backend.closeSection('section_run', run_gIndex)
            backend.finishedParsingSession("ParseSuccess", None)
            outF.write("]")



# Copyright 2015-2018 Ask Hjorth Larsen, Fawzi Mohamed, Ankit Kariryaa
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from __future__ import print_function
from subprocess import Popen, PIPE


octtype2nomadtype = {'float': 'f',
                     'integer': 'C', # Integers are really names, generally!
                     'logical': 'b',
                     'block': 'C',
                     'string': 'C',
                     'flag': 'C'}


json_header = """{
  "type": "nomad_meta_info_1_0",
  "description": "autogenerated nomad meta info for octopus parser",
  "dependencies": [ {
      "relativePath": "common.nomadmetainfo.json"
    }],
  "metaInfos": [ %(info)s]
}"""


json_template = """{
      "description": "%(description)s",
      "dtypeStr": "%(dtypeStr)s",
      "name": "%(name)s",
      "repeats": false,
      "shape": [],
      "superNames": [
        "%(supername)s"
      ]
    }"""

json_section_template = """{
      "description": "%(description)s",
      "kindStr": "type_abstract_document_content",
      "name": "%(name)s",
      "superNames": [
        "section_run"
      ]
    }"""



class OctInputVar:
    def __init__(self, name, octtype, section):
        self.name = name
        self.octtype = octtype
        self.section = section

    def _get_json(self, role, supername):
        description = (r'Octopus %s \"%s\" of type \"%s\" in section \"%s\"'
                       % (role, self.name, self.octtype, self.section))
        return json_template % dict(name='%s_%s' % (supername, self.name),
                                    dtypeStr=octtype2nomadtype[self.octtype],
                                    supername=supername,
                                    description=description)

    def get_input_json(self):
        return self._get_json('input parameter', 'x_octopus_input')

    def get_parserlog_json(self):
        return self._get_json('parser log entry', 'x_octopus_parserlog')


def main():
    searchproc = Popen(['oct-help', '--search', ''],
                       stdout=PIPE)

    def normalize(name):
        return var.name.lower() + '.'  # The '.' fiddles with EOL ordering

    oct_vars = {}

    def add(var):
        normalized_name = normalize(var.name)
        if normalized_name not in oct_vars:
            oct_vars[normalized_name] = var

    for line in searchproc.stdout.readlines():
        name = line.strip()
        printproc = Popen(('oct-help --print %s' % name).split(),
                          stdout=PIPE)
        fd = printproc.stdout

        # Example of output from oct-help:
        # ------------------
        #Variable: Spacing
        #Type:     float
        #Section:  Mesh
        #Description:
        #  The spacing between the points in the mesh. This controls the
        #  quality of the discretization: smaller spacing gives more precise
        #  results but increased computational cost.
        #  (.......)
        #------------------

        allowed_vardata = set(['Variable', 'Type', 'Default', 'Section'])

        vardata = {}
        for line in fd:
            title, remainder = line.split(':', 1)
            if title == 'Description':
                break
            vardata[title] = remainder.strip()

        var = OctInputVar(vardata.get('Variable'),
                          vardata.get('Type'),
                          vardata.get('Section'))

        add(var)

    itokens = []
    ptokens = []
    for key in sorted(oct_vars):
        var = oct_vars[key]
        itokens.append(var.get_input_json())
        ptokens.append(var.get_parserlog_json())

    itokens.append(json_section_template
                   % dict(description='section describing Octopus input '
                          'parameters',
                          name='x_octopus_input'))
    ptokens.append(json_section_template
                   % dict(description='section describing Octopus inputfile '
                          'parser log output',
                          name='x_octopus_parserlog'))

    print(json_header % dict(info=', '.join(itokens + ptokens)))

if __name__ == '__main__':
    main()



from builtins import object
from nomadcore.simple_parser import mainFunction, CachingLevel
from nomadcore.simple_parser import SimpleMatcher as SM
from nomadcore.local_meta_info import loadJsonFile, InfoKindEl
import os, sys, json, logging



################################################################
# This is the subparser for the WIEN2k input file (.in0)
################################################################

# Copyright 2016-2018 Daria M. Tomecka, Fawzi Mohamed
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Daria M. Tomecka"
__maintainer__ = "Daria M. Tomecka"
__email__ = "tomeckadm@gmail.com;"
__date__ = "15/05/2017"

class Wien2kIn0Context(object):
    """context for wien2k In0 parser"""

    def __init__(self):
        self.parser = None

    def initialize_values(self):
        """allows to reset values if the same superContext is used to parse different files"""
        pass

    def startedParsing(self, path, parser):
        """called when parsing starts"""
        self.parser = parser
        # allows to reset values if the same superContext is used to parse different files
        self.initialize_values()

    def onClose_x_wien2k_section_XC(self, backend, gIndex, section):
        xc_index = section["x_wien2k_indxc"]   #[0]
        #logging.error("winsectxc: %s -> %s", section, xc_index)
        if not xc_index:
            xc_index = ["XC_PBE"]
        xc_map_legend = {

            '5': ['LDA_C_PW_RPA'],
            'XC_LDA': ['LDA_X_2D'],

            '13': ['GGA_X_PBE', 'GGA_C_PBE'],
            'XC_PBE':['GGA_X_PBE', 'GGA_C_PBE'],

            '19': ['GGA_X_PBE_SOL', 'GGA_C_PBE_SOL'],
            'XC_PBESOL': ['GGA_X_PBE_SOL', 'GGA_C_PBE_SOL'],

            '11': ['GGA_X_WC'],
            'XC_WC': ['GGA_X_WC'],

            '17': ['GGA_X_PW91'],
            'EC_PW91': ['GGA_X_PW91'],
            'VC_PW91': ['GGA_X_PW91'],

            '28': ['MGGA_X_TB09'],
            'XC_MBJ': ['MGGA_X_TB09'],

            '29': ['MGGA_C_REVTPSS, GGA_C_REGTPSS'],
            'XC_REVTPSS': ['MGGA_C_REVTPSS, GGA_C_REGTPSS'],

            '24': ['GGA_X_B88', 'GGA_C_LYP'],
            'EX_B88': ['GGA_X_B88'],
            'VX_B88': ['GGA_X_B88'],
            'EC_LYP': ['GGA_C_LYP'],
            'VC_LYP': ['GGA_C_LYP'],

            '18': ['HYB_GGA_XC_B3PW91'],
            'XC_B3PW91': ['HYB_GGA_XC_B3PW91'],

            '27': ['MGGA_X_TPSS','MGGA_C_TPSS'],
            'XC_TPSS': ['MGGA_X_TPSS','MGGA_C_TPSS'],

            '46':['GGA_X_HTBS'],
            'XC_HTBS': ['GGA_X_HTBS'],

            '47': ['HYB_GGA_XC_B3LYP'],
            'XC_B3LYP': ['HYB_GGA_XC_B3LYP'],


            #        51: ['-'],
            #        EX_SLDA:
            #        VX_SLDA:

            #        52: ['-'],
            #        EX_SPBE:
            #        VX_SPBE:

            #        53: ['-'],
            #        EX_SWC:
            #        VX_SWC:

            #        54: ['-'],
            #        EX_SPBESOL:
            #        VX_SPBESOL:

            #        55: ['-'],
            #        EX_SB88:
            #        VX_SB88:

            '6': ['HF_X'],
            'EX_LDA': ['HF_X'],
            'VX_LDA': ['HF_X']
        }

        # Push the functional string into the backend
        xc_map_legend = xc_map_legend.get(xc_index[0])
        if not xc_map_legend:
            raise Exception("Unhandled xc functional %s found" % xc_index)


        for xc_name in xc_map_legend:
            #  for xc_name in xc_map_legend[xc_index]:
            s = backend.openSection("section_XC_functionals")
            backend.addValue("XC_functional_name", xc_name)
            backend.closeSection("section_XC_functionals", s)

# description of the input
def buildIn0Matchers():
    return SM(
        name = 'root',
        weak = True,
        startReStr = "",
        sections = ["section_run", "section_method"],
        subMatchers = [
            #        SM(name = 'systemName',
            #          startReStr = r"(?P<x_wien2k_system_nameIn>.*)"),
            SM(r"(?P<x_wien2k_switch>\w*)\s*(?P<x_wien2k_indxc>\w*)\s*.*",sections = ['x_wien2k_section_XC']),
            SM(r"\s*(?P<x_wien2k_ifft_x>[0-9]+)\s*(?P<x_wien2k_ifft_y>[0-9]+)\s*(?P<x_wien2k_ifft_z>[0-9]+)\s*(?P<x_wien2k_ifft_factor>[0-9.]+)\s*(?P<x_wien2k_iprint>[0-9]+).*")
        ])



def get_cachingLevelForMetaName(metaInfoEnv, CachingLvl):
    """Sets the caching level for the metadata.

    Args:
    metaInfoEnv: metadata which is an object of the class InfoKindEnv in nomadcore.local_meta_info.py.
    CachingLvl: Sets the CachingLevel for the sections k_band, run, and single_configuration_calculation.
    This allows to run the parser without opening new sections.

    Returns:
    Dictionary with metaname as key and caching level as value.
    """
    # manually adjust caching of metadata
    cachingLevelForMetaName = {
        'section_run': CachingLvl,
        'section_method': CachingLvl
    }
    return cachingLevelForMetaName

# loading metadata from nomad-meta-info/meta_info/nomad_meta_info/fhi_aims.nomadmetainfo.json



"""Modified JSON API v1.0 for OPTIMADE API"""
# pylint: disable=no-self-argument,no-name-in-module
from pydantic import Field, root_validator, BaseModel, AnyHttpUrl, AnyUrl, EmailStr
from typing import Optional, Union, List

from datetime import datetime

from . import jsonapi


__all__ = (
    "ResponseMetaQuery",
    "Provider",
    "ImplementationMaintainer",
    "Implementation",
    "ResponseMeta",
    "OptimadeError",
    "Success",
    "Warnings",
    "BaseRelationshipMeta",
    "BaseRelationshipResource",
    "Relationship",
)


class OptimadeError(jsonapi.Error):
    """detail MUST be present"""

    detail: str = Field(
        ...,
        description="A human-readable explanation specific to this occurrence of the problem.",
    )


class Warnings(OptimadeError):
    """OPTIMADE-specific warning class based on OPTIMADE-specific JSON API Error.
    From the specification:

        A warning resource object is defined similarly to a JSON API
        error object, but MUST also include the field type, which MUST
        have the value "warning". The field detail MUST be present and
        SHOULD contain a non-critical message, e.g., reporting
        unrecognized search attributes or deprecated features.

    Note: Must be named "Warnings", since "Warning" is a built-in Python class.
    """

    type: str = Field(
        "warning", const=True, description='Warnings must be of type "warning"'
    )

    @root_validator(pre=True)
    def status_must_not_be_specified(cls, values):
        if values.get("status", None) is not None:
            raise ValueError("status MUST NOT be specified for warnings")
        return values


class ResponseMetaQuery(BaseModel):
    """ Information on the query that was requested. """

    representation: str = Field(
        ...,
        description="a string with the part of the URL that follows the base URL. Example: '/structures?'",
    )


class Provider(BaseModel):
    """Information on the database provider of the implementation."""

    name: str = Field(..., description="a short name for the database provider")

    description: str = Field(
        ..., description="a longer description of the database provider"
    )

    prefix: str = Field(
        ..., description="database-provider-specific prefix as found in " "Appendix 1."
    )

    homepage: Optional[Union[AnyHttpUrl, jsonapi.Link]] = Field(
        None,
        description="a [JSON API links object](http://jsonapi.org/format/1.0#document-links) "
        "pointing to homepage of the database provider, either "
        "directly as a string, or as a link object.",
    )

    index_base_url: Optional[Union[AnyHttpUrl, jsonapi.Link]] = Field(
        None,
        description="a [JSON API links object](http://jsonapi.org/format/1.0#document-links) "
        "pointing to the base URL for the `index` meta-database as "
        "specified in Appendix 1, either directly as a string, or "
        "as a link object.",
    )


class ImplementationMaintainer(BaseModel):
    """Details about the maintainer of the implementation"""

    email: EmailStr = Field(..., description="the maintainer's email address")


class Implementation(BaseModel):
    """Information on the server implementation"""

    name: Optional[str] = Field(None, description="name of the implementation")

    version: Optional[str] = Field(
        None, description="version string of the current implementation"
    )

    source_url: Optional[AnyUrl] = Field(
        None,
        description="URL of the implementation source, either downloadable archive or version control system",
    )

    maintainer: Optional[ImplementationMaintainer] = Field(
        None,
        description="A dictionary providing details about the maintainer of the implementation.",
    )


class ResponseMeta(jsonapi.Meta):
    """
    A [JSON API meta member](https://jsonapi.org/format/1.0#document-meta)
    that contains JSON API meta objects of non-standard
    meta-information.

    OPTIONAL additional information global to the query that is not
    specified in this document, MUST start with a
    database-provider-specific prefix.
    """

    query: ResponseMetaQuery = Field(
        ..., description="information on the query that was requested"
    )

    api_version: str = Field(
        ...,
        description="a string containing the version of the API "
        "implementation, e.g. v0.9.5",
    )

    time_stamp: datetime = Field(
        ...,
        description="a string containing the date and time at which the query was exexcuted",
    )

    data_returned: int = Field(
        ...,
        description="an integer containing the number of data objects "
        "returned for the query.",
        ge=0,
    )

    more_data_available: bool = Field(
        ..., description="`false` if all data has been returned, and `true` " "if not."
    )

    provider: Provider = Field(
        ..., description="information on the database provider of the implementation."
    )

    data_available: Optional[int] = Field(
        None,
        description="an integer containing the total number of data "
        "objects available in the database",
    )

    last_id: Optional[str] = Field(
        None, description="a string containing the last ID returned"
    )

    response_message: Optional[str] = Field(
        None, description="response string from the server"
    )

    implementation: Optional[Implementation] = Field(
        None, description="a dictionary describing the server implementation"
    )

    warnings: Optional[List[Warnings]] = Field(
        None,
        description="List of warning resource objects representing non-critical errors or warnings. "
        "A warning resource object is defined similarly to a JSON API error object, but MUST also include the field type, "
        'which MUST have the value "warning". The field detail MUST be present and SHOULD contain a non-critical message, '
        "e.g., reporting unrecognized search attributes or deprecated features. The field status, representing a HTTP "
        "response status code, MUST NOT be present for a warning resource object. This is an exclusive field for error resource objects.",
        uniqueItems=True,
    )


class Success(jsonapi.Response):
    """errors are not allowed"""

    meta: Optional[ResponseMeta] = Field(
        None, description="A meta object containing non-standard information"
    )

    @root_validator(pre=True)
    def either_data_meta_or_errors_must_be_set(cls, values):
        """Overwriting the existing validation function, since 'errors' MUST NOT be set"""
        required_fields = ("data", "meta")
        if not any(values.get(field) for field in required_fields):
            raise ValueError(
                f"At least one of {required_fields} MUST be specified in the top-level response"
            )

        # errors MUST be skipped
        if values.get("errors", None) is not None:
            raise ValueError("'errors' MUST be skipped for a successful response")

        return values


class BaseRelationshipMeta(jsonapi.Meta):
    """Specific meta field for base relationship resource"""

    description: str = Field(
        ..., description="OPTIONAL human-readable description of the relationship"
    )


class BaseRelationshipResource(jsonapi.BaseResource):
    """Minimum requirements to represent a relationship resource"""

    meta: Optional[BaseRelationshipMeta] = Field(
        None,
        description="Relationship meta field. MUST contain 'description' if supplied.",
    )


class Relationship(jsonapi.Relationship):
    """Similar to normal JSON API relationship, but with addition of OPTIONAL meta field for a resource"""

    data: Optional[
        Union[BaseRelationshipResource, List[BaseRelationshipResource]]
    ] = Field(None, description="Resource linkage", uniqueItems=True)



# Copyright 2018 Markus Scheidgen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an"AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numpy as np
import re
import mmap
import logging


class Quantity:
    def __init__(self, name, re_pattern, str_operation=None, unit=None, dtype=None, comment=None):
        self.name = name
        self.re_pattern = re.compile(re_pattern.encode())
        self.unit = unit
        self.shape = None
        self.dtype = dtype
        self._value = None
        self._str_operation = str_operation
        self._comment = comment

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, val_in):

        def convert_value(val):
            if self._comment is not None:
                if val.strip()[0] == self._comment:
                    return

            if self._str_operation is not None:
                val = self._str_operation(val)

            else:
                val = val.strip().split() if isinstance(val, str) else val
                val = val[0] if len(val) == 1 else val

            def _convert(val):
                if isinstance(val, str):
                    if self.dtype is None:
                        if val.isdecimal():
                            val = int(val)
                        else:
                            try:
                                val = float(val)
                            except Exception:
                                pass

                    self.shape = []
                    return val

                elif type(val) in [np.ndarray, list]:
                    try:
                        dtype = float if self.dtype is None else self.dtype
                        val = np.array(val, dtype=dtype)
                        self.dtype = dtype
                        if np.all(np.mod(val, 1) == 0):
                            val = np.array(val, dtype=int)
                            self.dtype = int
                        self.shape = list(np.shape(val))

                    except Exception:
                        val = [_convert(v) for v in val]
                        self.dtype = None

                    return val

                else:
                    self.dtype = type(val)
                    self.shape = []
                    return val

            val = _convert(val)

            return val

        self._value = None if val_in is None else [convert_value(val) for val in val_in]

    def to_SI(self):
        pass


class UnstructuredTextFileParser:
    def __init__(self, mainfile, quantities, logger=None):
        self._mainfile = mainfile
        self.quantities = quantities
        self._file_mmap = None
        self.logger = logger if logger else logging

    @property
    def quantities(self):
        return self._quantities

    @quantities.setter
    def quantities(self, val):
        self._quantities = val
        self.quantities_mapping = {val[idx].name: idx for idx in range(len(val))}

    @property
    def file_mmap(self):
        if self._file_mmap is None:
            with open(self.mainfile) as f:
                self._file_mmap = mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ)
        return self._file_mmap

    @property
    def mainfile(self):
        return self._mainfile

    @mainfile.setter
    def mainfile(self, val):
        for quantity in self.quantities:
            quantity.value = None

        self._file_mmap = None
        self._mainfile = val

    def __getitem__(self, key):
        idx = self.quantities_mapping.get(key, None)
        if idx is None:
            return

        value = self.quantities[idx].value
        if value is None:
            self.parse(key)

        return self.quantities[idx].value

    def __setitem__(self, key, val):
        idx = self.quantities_mapping.get(key, None)
        if idx is None:
            return

        self.quantities[idx].value = val

    def keys(self):
        return self.quantities_mapping.keys()

    def items(self):
        for key in self.keys():
            yield key, self[key]

    def parse(self, key=None):
        if isinstance(key, str):
            key = [key]

        for quantity in self.quantities:
            if key is None or quantity.name in key:
                value = []
                for res in quantity.re_pattern.finditer(self.file_mmap):
                    value.append(''.join([group.decode() for group in res.groups() if group]))
                if not value:
                    continue

                try:
                    quantity.value = value
                except Exception:
                    self.logger.warn('Error setting value for %s ' % quantity.name)
                    pass



# Copyright 2016-2018 Carl Poelking, Fawzi Mohamed, Ankit Kariryaa
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from builtins import range
import os
import sys
import re
import json
import logging
import numpy as np

from nomadcore.local_meta_info import loadJsonFile, InfoKindEl
from contextlib import contextmanager

from dlpolyparser.libDlPolyParser import *

try:
    from dlpolyparser.libMomo import osio, endl, flush
    # osio.ConnectToFile('parser.osio.log')
    # green = osio.mg
    osio = endl = flush = None
    green = None
except:
    osio = endl = flush = None
    green = None


# parser_info = {
#     "name": "parser-dl-poly",
#     "version": "0.0",
#     "json": "../../../../nomad-meta-info/meta_info/nomad_meta_info/dl_poly.nomadmetainfo.json"
# }

# LOGGING
def log(msg, highlight=None, enter=endl):
    if osio:
        if highlight==None: hightlight = osio.ww
        osio << highlight << msg << enter
    return

# CONTEXT GUARD
@contextmanager
def open_section(p, name):
    gid = p.openSection(name)
    yield gid
    p.closeSection(name, gid)

def push(jbe, terminal, key1, fct=lambda x: x.As(), key2=None, conv=None):
    if key2 == None: key2 = key1
    value =  fct(terminal[key2])
    if conv:
        jbe.addValue(key1, value*conv)
    else:
        jbe.addValue(key1, value)
    return value

def push_array(jbe, terminal, key1, fct=lambda x: x.As(), key2=None, conv=None):
    if key2 == None: key2 = key1
    value =  np.asarray(fct(terminal[key2]))
    if conv:
        jbe.addArrayValues(key1, value*conv)
    else:
        jbe.addArrayValues(key1, value)
    return value

def push_value(jbe, value, key, conv=None):
    if conv:
        jbe.addValue(key, value*conv)
    else:
        jbe.addValue(key, value)
    return value

def push_array_values(jbe, value, key, conv=None):
    if conv:
        jbe.addArrayValues(key, value*conv)
    else:
        jbe.addArrayValues(key, value)
    return value


# class DlPolyParser():
#    """ A proper class envolop for running this parser from within python. """
#    def __init__(self, backend, **kwargs):
#        self.backend_factory = backend

#     from unittest.mock import patch
#     logging.info('dl-poly parser started')
#     logging.getLogger('nomadcore').setLevel(logging.WARNING)
#     backend = self.backend_factory(metaInfoEnv)
#     # Rename parser
#     def parse(self, mainfile):


logger = logging.getLogger("nomad.DLPolyParser")


class DlPolyParserWrapper():
    """ A proper class envolop for running this parser using Noamd-FAIRD infra. """
    def __init__(self, backend, **kwargs):
        self.backend_factory = backend

    def parse(self, mainfile):
        logging.info('dl-poly parser started')
        logging.getLogger('nomadcore').setLevel(logging.WARNING)
        backend = self.backend_factory("dl_poly.nomadmetainfo.json")
        # Call the old parser without a class.
        parserInfo = {'name': 'dl_poly-parser', 'version': '0.0'}
        backend = parse_without_class(mainfile, backend, parserInfo)
        return backend


def parse_without_class(output_file_name, backend, parser_info):
    """ Parse method to parse mainfile and write output to backend."""
    jbe = backend
    jbe.startedParsingSession(output_file_name, parser_info)

    base_dir = os.path.dirname(os.path.abspath(output_file_name))

    # PARSE CONTROLS ...
    ctrl_file_name = os.path.join(base_dir, 'CONTROL')
    terminal_ctrls = DlPolyControls(osio)
    terminal_ctrls.ParseControls(ctrl_file_name)

    # PARSE OUTPUT / TOPOLOGY ...
    output_file_name = os.path.join(base_dir, 'OUTPUT')
    terminal = DlPolyParser(osio)
    terminal.ParseOutput(output_file_name)

    # PARSE CONFIG ...
    cfg_file_name = os.path.join(base_dir, 'CONFIG')
    terminal_cfg = DlPolyConfig(osio)
    terminal_cfg.ParseConfig(cfg_file_name)

    # PARSE TRAJECTORY
    trj_file_name = os.path.join(base_dir, 'HISTORY')
    terminal_trj = DlPolyTrajectory(osio)
    terminal_trj.ParseTrajectory(trj_file_name, terminal.step_data, terminal.time_data)

    # SUMMARIZE KEY-TABLE DEFAULTS ...
    terminal.SummarizeKeyDefaults()
    terminal.topology.SummarizeKeyDefaults()
    terminal_ctrls.SummarizeKeyDefaults()
    terminal_cfg.SummarizeKeyDefaults()
    terminal_trj.SummarizeKeyDefaults()

    # ABBREVIATE ...
    ctr = terminal_ctrls
    out = terminal
    top = terminal.topology
    cfg = terminal_cfg
    trj = terminal_trj
    terminals = [ctr, out, top, cfg, trj]

    # ESTABLISH (ENERGY) UNITS
    unit_energy = top['energy_units'].As(str).lower().replace(' ','').replace('/','_').replace('(','_').replace(')','')
    UNITCONV_DLPOLY_TO_SI['energy'] = UNITCONV_DLPOLY_CUSTOM_TO_SI['energy_%s' % unit_energy]

    # ofs = open('parser.keys.log', 'w')
    # for quantity, conv in UNITCONV_DLPOLY_TO_SI.items():
    #     ofs.write('Unit [%s] = %e SI\n' % (quantity, conv))
    # ofs.write('\n')
    # for t in terminals:
    #     keys = sorted(t.data.keys())
    #     for key in keys:
    #         ofs.write('[%s] %s = %s\n' % (t.logtag, key, t[key].As(str)))
    #     ofs.write('\n')
    # ofs.close()

    # PUSH TO BACKEND
    with open_section(jbe, 'section_run') as gid_run:
        push(jbe, out, 'program_name')
        push(jbe, out, 'program_version')
        #push(jbe, out, 'program_info', key2='program_version_date')

        # TOPOLOGY SECTION
        with open_section(jbe, 'section_topology') as gid_top:
            # Cross-referencing is done on-the-fly (as gid's become available)
            # a) <molecule_to_molecule_type> :         shape=(number_of_topology_molecules, [<gid>])
            # b) <atom_in_molecule_to_atom_type_ref> : shape=(number_of_atoms_in_molecule, [<gid>])
            # c) <atom_to_molecule> :                  shape=(number_of_topology_atoms, [<molidx>, <atomidx>])
            push(jbe, top, 'number_of_topology_molecules', lambda s: s.As(int))
            push(jbe, top, 'number_of_topology_atoms', lambda s: s.As(int))
            # Atom types
            mol_type_atom_type_id_to_atom_type_gid = {}
            for mol in top.molecules:
                mol_name = mol['molecule_type_name'].As()
                mol_type_atom_type_id_to_atom_type_gid[mol_name] = {}
                for atom in mol.atoms:
                    # Add type
                    with open_section(jbe, 'section_atom_type') as gid_atom:
                        atom_id = atom['atom_id'].As(int)
                        mol_type_atom_type_id_to_atom_type_gid[mol_name][atom_id] = gid_atom
                        push(jbe, atom, 'atom_type_name', lambda s: s.As(), 'atom_name')
                        push(jbe, atom, 'atom_type_mass', lambda s: s.As(float), 'atom_mass', conv=UNITCONV_DLPOLY_TO_SI['mass'])
                        push(jbe, atom, 'atom_type_charge', lambda s: s.As(float), 'atom_charge', conv=UNITCONV_DLPOLY_TO_SI['charge'])
            # Molecule types
            molecule_type_name_to_type_gid = {}
            for mol in top.molecules:
                mol_name = mol['molecule_type_name'].As()
                # Extract references of atoms to atom types
                atom_type_id_to_atom_type_gid = mol_type_atom_type_id_to_atom_type_gid[mol_name]
                atom_gid_list = []
                for atom in mol.atoms:
                    atom_id = atom['atom_id'].As(int)
                    atom_gid_list.append(atom_type_id_to_atom_type_gid[atom_id])
                # Add molecule
                with open_section(jbe, 'section_molecule_type') as gid_mol:
                    molecule_type_name_to_type_gid[mol['molecule_type_name'].As()] = gid_mol
                    push(jbe, mol, 'molecule_type_name')
                    push(jbe, mol, 'number_of_atoms_in_molecule', lambda s: s.As(int))

                    push_array(jbe, mol, 'atom_in_molecule_name')
                    push_array(jbe, mol, 'atom_in_molecule_charge', conv=UNITCONV_DLPOLY_TO_SI['charge'])
                    push_array_values(jbe, np.asarray(atom_gid_list), 'atom_in_molecule_to_atom_type_ref')
            # Global molecule type map
            molecule_to_molecule_type = []
            for mol in top.molecules:
                type_name_this_mol = mol['molecule_type_name'].As()
                type_gid_this_mol = molecule_type_name_to_type_gid[type_name_this_mol]
                n_this_mol = mol['number_of_molecules'].As(int)
                for i in range(n_this_mol):
                    molecule_to_molecule_type.append(type_gid_this_mol)
            push_array_values(jbe, np.asarray(molecule_to_molecule_type), 'molecule_to_molecule_type_map')

            # Global atom map
            atoms_to_molidx_atomidx = []
            molidx = 0
            for mol in top.molecules:
                n_mol = mol['number_of_molecules'].As(int)
                for i in range(n_mol):
                    atomidx = 0
                    for atom in mol.atoms:
                        molidx_atomidx = [ molidx, atomidx ]
                        atoms_to_molidx_atomidx.append(molidx_atomidx)
                        atomidx += 1
                    molidx += 1
            push_array_values(jbe, np.asarray(atoms_to_molidx_atomidx), 'atom_to_molecule')

        # SAMPLING-METHOD SECTION
        sec_sampling_method_ref = None
        with open_section(jbe, 'section_sampling_method') as gid_sec_sampling_method:
            sec_sampling_method_ref = gid_sec_sampling_method
            # Ensemble
            ensemble = push(jbe, out, 'ensemble_type', lambda s: s.As().split()[0].upper())
            # Method
            push(jbe, out, 'sampling_method')
            push(jbe, out, 'x_dl_poly_integrator_type')
            push(jbe, out, 'x_dl_poly_integrator_dt', lambda s: s.As(float))
            push(jbe, out, 'x_dl_poly_number_of_steps_requested', lambda s: s.As(int))
            # Coupling
            if 'T' in ensemble:
                push(jbe, out, 'x_dl_poly_thermostat_target_temperature', lambda s: s.As(float), conv=UNITCONV_DLPOLY_TO_SI['temperature'])
                push(jbe, out, 'x_dl_poly_thermostat_tau', lambda s: s.As(float))
                pass
            if 'P' in ensemble:
                push(jbe, out, 'x_dl_poly_barostat_target_pressure', lambda s: s.As(float), conv=UNITCONV_DLPOLY_TO_SI['pressure_katm'])
                push(jbe, out, 'x_dl_poly_barostat_tau', lambda s: s.As(float), conv=UNITCONV_DLPOLY_TO_SI['time'])
                pass
            pass

        # METHOD SECTION
        sec_method_ref = None
        with open_section(jbe, 'section_method') as gid_sec_method:
            sec_method_ref = gid_sec_method
            push_value(jbe, 'force_field', 'calculation_method')

        # TODO Store state variables in frames/system description (temperature, pressure)
        # TODO Store interactions

        # SYSTEM DESCRIPTION
        refs_system_description = []
        all_frames = [cfg] + trj.frames # <- Initial config + trajectory
        for frame in all_frames:
            with open_section(jbe, 'section_system') as gid:
                refs_system_description.append(gid)
                # Configuration core
                atom_labels = np.array(
                    [ atom['atom_name'].As().replace('+','').replace('-','') for atom in frame.atoms ])
                push_array_values(jbe, atom_labels, 'atom_labels')
                # push_array_values(jbe, atom_species, 'atom_atom_numbers')
                push_array_values(jbe, frame.position_matrix, 'atom_positions', conv=UNITCONV_DLPOLY_TO_SI['length'])
                push_array_values(jbe, frame.box_matrix, 'simulation_cell', conv=UNITCONV_DLPOLY_TO_SI['length'])
                push_array_values(jbe, frame.pbc_booleans, 'configuration_periodic_dimensions')
                if frame.has_velocities:
                    push_array_values(jbe, frame.velocity_matrix, 'atom_velocities', conv=UNITCONV_DLPOLY_TO_SI['length']/UNITCONV_DLPOLY_TO_SI['time'])
                if frame.has_forces:
                    # TODO Wouldn't it be simpler if forces were added here?
                    pass
                pass

        # SINGLE CONFIGURATIONS
        refs_single_configuration = []
        i_frame = -1
        for frame in all_frames:
            i_frame += 1
            with open_section(jbe, 'section_single_configuration_calculation') as gid:
                refs_single_configuration.append(gid)
                # Reference system description section
                ref_system = refs_system_description[i_frame]
                push_value(jbe, ref_system, 'single_configuration_calculation_to_system_ref')
                # Energy total
                if frame.has_energy_total:
                    push_value(jbe, frame.energy_total, 'energy_total', conv=UNITCONV_DLPOLY_TO_SI['energy'])
                # Forces
                if frame.has_forces:
                    push_array_values(jbe, frame.force_matrix, 'atom_forces', conv=UNITCONV_DLPOLY_TO_SI['mass']*UNITCONV_DLPOLY_TO_SI['length']/UNITCONV_DLPOLY_TO_SI['time']**2)
                # Method reference
                push_value(jbe, sec_method_ref, 'single_configuration_to_calculation_method_ref')
                pass

        # FRAME-SEQUENCE SECTION
        with open_section(jbe, 'section_frame_sequence'):
            push_value(jbe, len(all_frames), 'number_of_frames_in_sequence')
            # Reference configurations and sampling method
            push_value(jbe, sec_sampling_method_ref, 'frame_sequence_to_sampling_ref')
            refs_config = np.array(refs_single_configuration)
            push_array_values(jbe, refs_config, 'frame_sequence_local_frames_ref')
            time_values = np.array([ frame['time_value'].As(float)*UNITCONV_DLPOLY_TO_SI['time'] for frame in trj.frames ])
            push_array_values(jbe, time_values, 'frame_sequence_time')
            pass

    jbe.finishedParsingSession("ParseSuccess", None)
    return jbe

if __name__ == '__main__':

    # CALCULATE PATH TO META-INFO FILE
    this_py_file = os.path.abspath(__file__)
    this_py_dirname = os.path.dirname(this_py_file)
    json_supp_file = parser_info["json"]
    meta_info_path = os.path.normpath(os.path.join(this_py_dirname, json_supp_file))

    # LOAD META-INFO FILE
    log("Meta-info from '%s'" % meta_info_path)
    meta_info_env, warns = loadJsonFile(
        filePath=meta_info_path,
        dependencyLoader=None,
        extraArgsHandling=InfoKindEl.ADD_EXTRA_ARGS,
        uri=None)

    output_file_name = sys.argv[1]
    parse(output_file_name)





# Copyright 2016-2018 Fawzi Mohamed, Lauri Himanen, Danio Brambila, Ankit Kariryaa, Henning Glawe
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import os
import logging
import gzip
import bz2
import lzma

from nomad.datamodel import EntryArchive

from .metainfo import m_env
from vaspparser.parser_vasprun import VasprunContext, XmlParser, parser_info
from nomad.parsing.parser import FairdiParser
from vaspparser.parser_outcar import VaspOutcarParser


class VASPParser(FairdiParser):
    def __init__(self):
        super().__init__(
            name='parsers/vasp', code_name='VASP', code_homepage='https://www.vasp.at/',
            mainfile_mime_re=r'(application/.*)|(text/.*)',
            mainfile_contents_re=(
                r'^\s*<\?xml version="1\.0" encoding="ISO-8859-1"\?>\s*'
                r'?\s*<modeling>'
                r'?\s*<generator>'
                r'?\s*<i name="program" type="string">\s*vasp\s*</i>'
                r'?'),
            supported_compressions=['gz', 'bz2', 'xz']
        )

    def parse(self, filepath, archive, logger=None):
        self._metainfo_env = m_env

        super_context = VasprunContext(logger=logger)

        parser = XmlParser(parser_info, super_context, metainfo_env=m_env)

        open_file = open
        if filepath.endswith('.gz'):
            open_file = gzip.open
        elif filepath.endswith('.bz2'):
            open_file = bz2.open
        elif filepath.endswith('.xz'):
            open_file = lzma.open

        parser.parse(os.path.abspath(filepath), open_file(filepath, 'rt'), archive)



# Copyright 2015-2018 Lauri Himanen, Fawzi Mohamed, Ankit Kariryaa
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from future import standard_library
standard_library.install_aliases()
from builtins import object
import os
import re
from io import StringIO
import logging
import pickle
import numpy as np
from nomadcore.baseclasses import AbstractBaseParser
import nomadcore.configurationreading
from cp2kparser.generic.inputparsing import metainfo_data_prefix, metainfo_section_prefix
from pint import UnitRegistry
import ase

ureg = UnitRegistry()
logger = logging.getLogger("nomad")


class CP2KInputParser(AbstractBaseParser):
    """Used to parse out a CP2K input file.

    CP2K offers a complete structure for the input in an XML file, which can be
    printed with the command cp2k --xml. This XML file has been preparsed into
    a native python object ('CP2KInput' class found in generic.inputparsing)
    and stored in a python pickle file. It e.g. contains all the default values
    that are often needed as they are used if the user hasn't specified a
    settings in the input. This XML file is used to get the default values
    because it is rather cumbersome to hard code them in the parser itself,
    especially if there will be lot's of them. Hard coded values will also be
    more error prone, and would have to be checked for each parser version.

    CP2K input supports including other input files and also
    supports variables. This is currently not supported, but may be added at
    some point.
    """
    def __init__(self, parser_context):
        """
        Attributes:
            input_tree: The input structure for this version of CP2K. The
                structure is already present, in this module it will be filled with
                data found from the input file.
            input_lines: List of preprocessed lines in the input. Here all the
                variables have been stated explicitly and the additional input files have
                been merged.
        """
        super(CP2KInputParser, self).__init__(parser_context)
        self.input_tree = None
        self.input_lines = None
        self.unit_mapping = {
            # Distance
            "BOHR": "bohr",
            "M": "m",
            "PM": "pm",
            "NM": "nm",
            "ANGSTROM": "angstrom",
            # Time
            "S": "s",
            "FS": "fs",
            "PS": "ps",
            "AU_T": "(hbar/hartree)",
            "WAVENUMBER_T": None,
        }

        #=======================================================================
        # Cached values
        self.cache_service.add("configuration_periodic_dimensions", single=False, update=False)
        self.cache_service.add("trajectory_format")
        self.cache_service.add("trajectory_unit")
        self.cache_service.add("velocity_format")
        self.cache_service.add("velocity_unit")
        self.cache_service.add("vel_add_last")
        self.cache_service.add("each_geo_opt")
        self.cache_service.add("traj_add_last")

    def parse(self, filepath):

        # Preprocess to spell out variables and to include stuff from other
        # files
        self.preprocess_input(filepath)

        # Gather the information from the input file
        self.fill_input_tree(filepath)

        # Parse everything in the input to cp2k specific metadata
        self.fill_metadata()

        # If the print level is low, parse the system geometry from the input
        # file. If keywords that manipulate the system are enabled, the
        # geometry is not parsed because the manipulations can be hard to
        # reproduce.
        print_level = self.input_tree.get_keyword("GLOBAL/PRINT_LEVEL")
        multiple_unit_cell = self.input_tree.get_keyword("FORCE_EVAL/SUBSYS/MULTIPLE_UNIT_CELL")
        top_multiple_unit_cell = self.input_tree.get_keyword("FORCE_EVAL/SUBSYS/TOPOLOGY/MULTIPLE_UNIT_CELL")
        centering = self.input_tree.get_section("FORCE_EVAL/SUBSYS/TOPOLOGY/CENTER_COORDINATES").section_parameter.get_value()
        center_point = self.input_tree.get_keyword("FORCE_EVAL/SUBSYS/TOPOLOGY/CENTER_COORDINATES/CENTER_POINT")
        if multiple_unit_cell is None:
            multiple_unit_cell = np.array([1, 1, 1])

        if print_level == "LOW" \
           and np.array_equal(multiple_unit_cell, [1, 1, 1]) \
           and np.array_equal(top_multiple_unit_cell, [1, 1, 1]):

            cell_section = self.input_tree.get_section("FORCE_EVAL/SUBSYS/CELL")
            a_obj = cell_section.get_keyword_object("A")
            b_obj = cell_section.get_keyword_object("B")
            c_obj = cell_section.get_keyword_object("C")
            a_val = a_obj.get_value()
            b_val = b_obj.get_value()
            c_val = c_obj.get_value()

            abc_obj = cell_section.get_keyword_object("ABC")
            angles_obj = cell_section.get_keyword_object("ALPHA_BETA_GAMMA")
            abc_val = abc_obj.get_value()
            angles_val = angles_obj.get_value()

            # Cell given as three vectors
            if a_val is not None and b_val is not None and c_val is not None:
                a_unit = a_obj.get_unit()
                b_unit = b_obj.get_unit()
                c_unit = c_obj.get_unit()
                a = (a_val * ureg(a_unit)).to("angstrom").magnitude
                b = (b_val * ureg(b_unit)).to("angstrom").magnitude
                c = (c_val * ureg(c_unit)).to("angstrom").magnitude
                cell = np.stack((a, b, c), axis=0)
                self.cache_service["simulation_cell"] = cell
                self.cache_service["lattice_vectors"] = cell
            # Cell given as lengths and angles
            elif abc_val is not None and angles_val is not None:
                abc_unit = abc_obj.get_unit()
                angles_unit = angles_obj.get_unit()
                abc = (abc_val * ureg(abc_unit)).to("angstrom").magnitude
                angles = (angles_val * ureg(angles_unit)).to("degree").magnitude
                parameters = np.hstack((abc, angles))
                cell = ase.geometry.cellpar_to_cell(parameters)
                self.cache_service["simulation_cell"] = cell
                self.cache_service["lattice_vectors"] = cell

            # Parse the coordinates from input file or xyz file
            coord = self.input_tree.get_section("FORCE_EVAL/SUBSYS/COORD")
            coord_filename = self.input_tree.get_keyword("FORCE_EVAL/SUBSYS/TOPOLOGY/COORD_FILE_NAME")
            coord_format = self.input_tree.get_keyword("FORCE_EVAL/SUBSYS/TOPOLOGY/COORD_FILE_FORMAT")
            scaled = coord.get_keyword("SCALED")
            coords = coord.default_keyword.value
            coord_unit = coord.get_keyword("UNIT")

            final_pos = None
            final_lab = None
            if len(coords) != 0:
                lines = coords.split("\n")
                pos = []
                labels = []
                for line in lines:
                    if len(line) != 0:
                        parts = line.split()
                        i_lab = parts[0]
                        i_x = float(parts[1])
                        i_y = float(parts[2])
                        i_z = float(parts[3])
                        labels.append(i_lab)
                        pos.append([i_x, i_y, i_z])

                final_pos = np.array(pos)
                final_lab = np.array(labels)

            elif coord_filename is not None:

                extension_map = {
                    "XYZ": "xyz",
                }
                coord_extension = extension_map.get(coord_format)
                if coord_extension is not None:

                    # We first remove the comment line from the XYZ file,
                    # because ASE cannot properly handle comments that have ASE
                    # specific data that is malformed.
                    abs_filename = self.file_service.get_absolute_path_to_file(coord_filename)
                    stringio = StringIO()
                    with open(abs_filename, "r") as fin:
                        for i_line, line in enumerate(fin):
                            if i_line == 1:
                                stringio.write("\n")
                            else:
                                stringio.write(line)
                    try:
                        atoms = ase.io.read(stringio, format="xyz")
                        final_pos = atoms.get_positions()
                        final_lab = atoms.get_chemical_symbols()
                    except StopIteration:
                        logger.error("Error in reading the structure file specified in the input file.")
                else:
                    logger.error(
                        "Could not read the structure file specified in the "
                        "input file because the file format is not supported."
                    )

            # Make the positions cartesian if they are scaled
            if final_pos is not None and final_lab is not None:
                if scaled == "T" or scaled == "TRUE":
                    atoms = ase.Atoms(
                        scaled_positions=final_pos,
                        symbols=final_lab,
                        cell=cell,
                    )
                    final_pos = atoms.get_positions()

                # Center the atoms if requested.
                if centering == "T" or centering == "TRUE":
                    atoms = ase.Atoms(
                        positions=final_pos,
                        symbols=final_lab,
                        cell=cell,
                    )
                    atoms.center(about=center_point)
                    final_pos = atoms.get_positions()

                coord_unit = self.get_pint_unit_string(coord_unit)
                final_pos = (final_pos * ureg(coord_unit)).to("angstrom").magnitude
                self.cache_service["atom_positions"] = final_pos
                self.cache_service["atom_labels"] = final_lab

        # Parse the used XC_functionals and their parameters
        xc = self.input_tree.get_section("FORCE_EVAL/DFT/XC/XC_FUNCTIONAL")
        if xc is not None:
            xc_list = []

            class XCFunctional(object):
                def __init__(self, name, weight=1, parameters=None):
                    self.name = name
                    self.weight = weight
                    self.parameters = parameters

            # First see if a functional has been specified in the section parameter
            section_parameter = xc.section_parameter.value
            if section_parameter is not None and section_parameter != "NO_SHORTCUT":

                if section_parameter == "BLYP":
                    xc_list.append(XCFunctional("GGA_X_B88"))
                    xc_list.append(XCFunctional("GGA_C_LYP"))

                elif section_parameter == "LDA" or section_parameter == "PADE":
                    xc_list.append(XCFunctional("LDA_XC_TETER93"))

                elif section_parameter == "PBE":
                    xc_list.append(XCFunctional("GGA_X_PBE"))
                    xc_list.append(XCFunctional("GGA_C_PBE"))

                elif section_parameter == "OLYP":
                    xc_list.append(XCFunctional("GGA_X_OPTX"))
                    xc_list.append(XCFunctional("GGA_C_LYP"))

                elif section_parameter == "HCTH120":
                    xc_list.append(XCFunctional("GGA_XC_HCTH_120"))

                elif section_parameter == "PBE0":
                    xc_list.append(XCFunctional("HYB_GGA_XC_PBEH"))

                elif section_parameter == "B3LYP":
                    xc_list.append(XCFunctional("HYB_GGA_XC_B3LYP"))

                elif section_parameter == "TPSS":
                    xc_list.append(XCFunctional("MGGA_X_TPSS"))
                    xc_list.append(XCFunctional("MGGA_C_TPSS"))

                else:
                    logger.warning("Unknown XC functional given in XC_FUNCTIONAL section parameter.")

            # Otherwise one has to look at the individual functional settings
            else:
                pbe = xc.get_subsection("PBE")
                if pbe is not None:
                    if pbe.accessed:
                        sp = pbe.get_section_parameter()
                        if sp == "T":
                            parametrization = pbe.get_keyword("PARAMETRIZATION", allow_default=True)
                            scale_x = pbe.get_keyword("SCALE_X", allow_default=True)
                            scale_c = pbe.get_keyword("SCALE_C", allow_default=True)
                            if parametrization == "ORIG":
                                xc_list.append(XCFunctional("GGA_X_PBE", scale_x))
                                xc_list.append(XCFunctional("GGA_C_PBE", scale_c))
                            elif parametrization == "PBESOL":
                                xc_list.append(XCFunctional("GGA_X_PBE_SOL", scale_x))
                                xc_list.append(XCFunctional("GGA_C_PBE_SOL", scale_c))
                            elif parametrization == "REVPBE":
                                xc_list.append(XCFunctional("GGA_X_PBE_R", scale_x))
                                xc_list.append(XCFunctional("GGA_C_PBE", scale_c))
                tpss = xc.get_subsection("TPSS")
                if tpss is not None:
                    if tpss.accessed:
                        sp = tpss.get_section_parameter()
                        if sp == "T":
                            scale_x = tpss.get_keyword("SCALE_X", allow_default=True)
                            scale_c = tpss.get_keyword("SCALE_C", allow_default=True)
                            xc_list.append(XCFunctional("MGGA_X_TPSS", scale_x))
                            xc_list.append(XCFunctional("MGGA_C_TPSS", scale_c))

            # Sort the functionals alphabetically by name
            xc_list.sort(key=lambda x: x.name)
            xc_summary = ""

            # For every defined functional, stream the information to the
            # backend and construct the summary string
            for i, functional in enumerate(xc_list):

                gId = self.backend.openSection("section_XC_functionals")
                self.backend.addValue("XC_functional_name", functional.name)
                self.backend.addValue("XC_functional_weight", functional.weight)
                if functional.parameters is not None:
                    pass
                self.backend.closeSection("section_XC_functionals", gId)

                if i != 0:
                    xc_summary += "+"
                xc_summary += "{}*{}".format(functional.weight, functional.name)
                if functional.parameters is not None:
                    xc_summary += ":{}".format()

            # Stream summary
            if xc_summary is not "":
                self.backend.addValue("XC_functional", xc_summary)

        #=======================================================================
        # Cell periodicity
        periodicity = self.input_tree.get_keyword("FORCE_EVAL/SUBSYS/CELL/PERIODIC")
        if periodicity is not None:
            periodicity = periodicity.upper()
            periodicity_list = ("X" in periodicity, "Y" in periodicity, "Z" in periodicity)
            self.cache_service["configuration_periodic_dimensions"] = np.asarray(periodicity_list)
        else:
            logger.warning("Could not determine cell periodicity from FORCE_EVAL/SUBSYS/CELL/PERIODIC")

        #=======================================================================
        # Single point force file name
        self.setup_force_file_name()

        #=======================================================================
        # Trajectory file name
        self.setup_trajectory_file_name()

        #=======================================================================
        # Trajectory file format
        self.cache_service["trajectory_format"] = self.input_tree.get_keyword("MOTION/PRINT/TRAJECTORY/FORMAT")
        self.cache_service["traj_add_last"] = self.input_tree.get_keyword("MOTION/PRINT/TRAJECTORY/ADD_LAST")
        traj_unit = self.input_tree.get_keyword("MOTION/PRINT/TRAJECTORY/UNIT")
        pint_traj_unit = self.get_pint_unit_string(traj_unit)
        self.cache_service["trajectory_unit"] = pint_traj_unit

        #=======================================================================
        # Velocity file format
        self.cache_service["velocity_format"] = self.input_tree.get_keyword("MOTION/PRINT/VELOCITIES/FORMAT")
        self.cache_service["vel_add_last"] = self.input_tree.get_keyword("MOTION/PRINT/VELOCITIES/ADD_LAST")
        vel_unit = self.input_tree.get_keyword("MOTION/PRINT/VELOCITIES/UNIT")
        pint_vel_unit = self.get_pint_unit_string(vel_unit)
        self.cache_service["velocity_unit"] = pint_vel_unit

        #=======================================================================
        # OLD: This information is retrieved from output file
        # See if some more exotic calculation is requested (e.g. MP2, DFT+U, GW, RPA)
        # Search for a WF_CORRELATION section
        # correlation = self.input_tree.get_section("FORCE_EVAL/DFT/XC/WF_CORRELATION")
        # method = "DFT"
        # if correlation.accessed:
            # method = correlation.get_keyword_value_raw("METHOD")
            # if method != "NONE":
                # # Can't really decide which method used (MP2, RPA, GW)
                # method = None

        # # Search for DFT+U settings
        # kinds = self.input_tree.get_section_list("FORCE_EVAL/SUBSYS/KIND")
        # for kind in kinds:
            # dft_u = kind.get_subsection("DFT_PLUS_U")
            # if dft_u.accessed:
                # method = "DFT+U"

        # self.cache_service["electronic_structure_method"] = method

        #=======================================================================
        # Stress tensor calculation method
        stress_tensor_method = self.input_tree.get_keyword("FORCE_EVAL/STRESS_TENSOR")
        if stress_tensor_method != "NONE":
            mapping = {
                "NUMERICAL": "Numerical",
                "ANALYTICAL": "Analytical",
                "DIAGONAL_ANALYTICAL": "Diagonal analytical",
                "DIAGONAL_NUMERICAL": "Diagonal numerical",
            }
            stress_tensor_method = mapping.get(stress_tensor_method)
            if stress_tensor_method is not None:
                self.backend.addValue("stress_tensor_method", stress_tensor_method)

    def normalize_x_cp2k_path(self, path):
        """The paths in CP2K input can be given in many ways. This function
        tries to normalize these forms into a valid path.
        """
        # Path is exactly as given
        if path.startswith("="):
            normalized_path = path[1:]
        # Path is relative, no project name added
        elif re.match(r"./", path):
            normalized_path = path
        # Path is relative, project name added
        else:
            project_name = self.input_tree.get_keyword("GLOBAL/PROJECT_NAME")
            if path:
                normalized_path = "{}-{}".format(project_name, path)
            else:
                normalized_path = project_name
        return normalized_path

    def get_pint_unit_string(self, cp2k_unit_string):
        """Translate the CP2K unit definition into a valid pint unit.
        """
        units = re.split('[\^\-\+\*\d]+', cp2k_unit_string)
        for unit in units:
            if unit == "":
                continue
            pint_unit = self.unit_mapping.get(unit.upper())
            if pint_unit is None:
                return None
            cp2k_unit_string = cp2k_unit_string.replace(unit, pint_unit)
        return cp2k_unit_string

    def setup_force_file_name(self):
        """Setup the force file path.
        """
        force_file = self.input_tree.get_keyword("FORCE_EVAL/PRINT/FORCES/FILENAME")
        extension = "xyz"
        if force_file is not None and force_file != "__STD_OUT__":
            normalized_path = self.normalize_x_cp2k_path(force_file)
            final_path = "{}-1_0.{}".format(normalized_path, extension)
            self.file_service.set_file_id(final_path, "force_file_single_point")

    def setup_trajectory_file_name(self):
        """Setup the trajectory file path.
        """
        traj_format = self.input_tree.get_keyword("MOTION/PRINT/TRAJECTORY/FORMAT")
        traj_filename = self.input_tree.get_keyword("MOTION/PRINT/TRAJECTORY/FILENAME")
        self.cache_service["each_geo_opt"] = self.input_tree.get_keyword("MOTION/PRINT/TRAJECTORY/EACH/GEO_OPT")
        if traj_filename is None:
            traj_filename = ""
        extension_map = {
            "XYZ": "xyz",
            "XMOL": "xyz",
            "ATOMIC": "xyz",
            "PDB": "pdb",
            "DCD": "dcd",
        }
        extension = extension_map.get(traj_format)
        if extension is None:
            logger.error("Unknown file format '{}' for CP2K trajectory file ".format(traj_format))
            return
        normalized_path = self.normalize_x_cp2k_path(traj_filename)
        final_path = "{}-pos-1.{}".format(normalized_path, extension)
        self.file_service.set_file_id(final_path, "trajectory")

    def fill_input_tree(self, file_path):
        """Parses a CP2K input file into an object tree.

        Return an object tree represenation of the input augmented with the
        default values and lone keyword values from the x_cp2k_input.xml file
        which is version specific. Keyword aliases are also mapped to the same
        data.

        The cp2k input is largely case-insensitive. In the input tree, we wan't
        only one standard way to name things, so all section names and section
        parameters will be transformed into upper case.

        To query the returned tree use the following functions:
            get_keyword("GLOBAL/PROJECT_NAME")
            get_parameter("GLOBAL/PRINT")
            get_default_keyword("FORCE_EVAL/SUBSYS/COORD")

        Args:
            : A string containing the contents of a CP2K input file. The
            input file can be stored as string as it isn't that big.

        Returns:
            The input as an object tree.
        """

        self.setup_version(self.parser_context.version_id)
        section_stack = []
        self.input_tree.root_section.accessed = True

        for line in self.input_lines:

            # Remove comments and whitespaces
            line = line.split('!', 1)[0].split('#', 1)[0].strip()

            # Skip empty lines
            if len(line) == 0:
                continue

            # Section ends
            if line.upper().startswith('&END'):
                section_stack.pop()
            # Section starts
            elif line[0] == '&':
                parts = line.split(' ', 1)
                name = parts[0][1:].upper()
                section_stack.append(name)

                # Form the path
                path = ""
                for index, item in enumerate(section_stack):
                    if index != 0:
                        path += '/'
                    path += item

                # Mark the section as accessed.
                self.input_tree.set_section_accessed(path)

                # Save the section parameters
                if len(parts) > 1:
                    self.input_tree.set_parameter(path, parts[1].strip().upper())

            # Ignore variables and includes that might still be here for some
            # reason
            elif line.upper().startswith('@'):
                continue

            # Contents (keywords, default keywords)
            else:
                split = line.split(None, 1)
                if len(split) <= 1:
                    keyword_value = ""
                else:
                    keyword_value = split[1]
                keyword_name = split[0].upper()
                self.input_tree.set_keyword(path + "/" + keyword_name, keyword_value, line)

    def fill_metadata(self):
        """Goes through the input data and pushes everything to the
        backend.
        """
        name_stack = []
        self.fill_metadata_recursively(self.input_tree.root_section, name_stack)

    def fill_metadata_recursively(self, section, name_stack):
        """Recursively goes through the input sections and pushes everything to the
        backend.
        """
        if not section.accessed:
            return

        if section.name == "CP2K_INPUT":
            path = "x_cp2k_section_input"
        else:
            name_stack.append(section.name)
            path = metainfo_section_prefix + "{}".format(".".join(name_stack))

        not_section_path = metainfo_data_prefix + "{}".format(".".join(name_stack))

        gid = self.backend.openSection(path)

        # Keywords
        for default_name in section.default_keyword_names:
            keywords = section.keywords.get(default_name)
            for keyword in keywords:
                if keyword.value is not None:
                    name = "{}.{}".format(not_section_path, keyword.default_name)
                    self.backend.addValue(name, keyword.value)

        # Section parameter
        section_parameter = section.section_parameter
        if section_parameter is not None:
            name = "{}.SECTION_PARAMETERS".format(not_section_path)
            if section_parameter.value is not None:
                self.backend.addValue(name, section_parameter.value)

        # Default keyword
        default_keyword = section.default_keyword
        if default_keyword is not None:

            name = "{}.DEFAULT_KEYWORD".format(not_section_path)
            self.backend.addValue(name, default_keyword.value)

        # Subsections
        for name, subsections in section.sections.items():
            for subsection in subsections:
                self.fill_metadata_recursively(subsection, name_stack)

        self.backend.closeSection(path, gid)

        if section.name != "CP2K_INPUT":
            name_stack.pop()

    def setup_version(self, version_number):
        """ The pickle file which contains preparsed data from the
        x_cp2k_input.xml is version specific. By calling this function before
        parsing the correct file can be found.
        """
        pickle_path = os.path.dirname(__file__) + "/input_data/cp2k_input_tree.pickle"
        input_tree_pickle_file = open(pickle_path, 'rb')
        self.input_tree = pickle.load(input_tree_pickle_file)

    def preprocess_input(self, filepath):
        """Preprocess the input file. Concatenate .inc files into the main
        input file and explicitly state all variables.
        """
        # Read the input file into memory. It shouldn't be that big so we can
        # do this easily
        input_lines = []
        with open(filepath, "r") as f:
            for line in f:
                input_lines.append(line.strip())

        # Merge include files to input
        extended_input = input_lines[:]  # Make a copy
        i_line = 0
        for line in input_lines:
            if line.startswith("@INCLUDE") or line.startswith("@include"):
                split = line.split(None, 1)
                includepath = split[1]
                basedir = os.path.dirname(filepath)
                filepath = os.path.join(basedir, includepath)
                filepath = os.path.abspath(filepath)
                if not os.path.isfile(filepath):
                    logger.warning("Could not find the include file '{}' stated in the CP2K input file. Continuing without it.".format(filepath))
                    continue

                # Get the content from include file
                included_lines = []
                with open(filepath, "r") as includef:
                    for line in includef:
                        included_lines.append(line.strip())
                    del extended_input[i_line]
                    extended_input[i_line:i_line] = included_lines
                    i_line += len(included_lines)
            i_line += 1

        # Gather the variable definitions
        variables = {}
        input_set_removed = []
        for i_line, line in enumerate(extended_input):
            if line.startswith("@SET") or line.startswith("@set"):
                components = line.split(None, 2)
                name = components[1]
                value = components[2]
                variables[name] = value
                logger.debug("Variable '{}' found with value '{}'".format(name, value))
            else:
                input_set_removed.append(line)

        # Place the variables
        variable_pattern = r"\$\{(\w+)\}|\$(\w+)"
        compiled = re.compile(variable_pattern)
        reserved = ("include", "set", "if", "endif")
        input_variables_replaced = []
        for line in input_set_removed:
            results = compiled.finditer(line)
            new_line = line
            offset = 0
            for result in results:
                options = result.groups()
                first = options[0]
                second = options[1]
                if first:
                    name = first
                elif second:
                    name = second
                if name in reserved:
                    continue
                value = variables.get(name)
                if not value:
                    logger.error("Value for variable '{}' not set.".format(name))
                    continue
                len_value = len(value)
                len_name = len(name)
                start = result.start()
                end = result.end()
                beginning = new_line[:offset+start]
                rest = new_line[offset+end:]
                new_line = beginning + value + rest
                offset += len_value - len_name - 1
            input_variables_replaced.append(new_line)

        self.input_lines = input_variables_replaced



from typing import Tuple
from optimade.server.config import CONFIG

__all__ = ("BaseResourceMapper",)


class BaseResourceMapper:
    """ Generic Resource Mapper that defines and performs the mapping
    between objects in the database and the resource objects defined by
    the specification.

    Attributes:
        ENDPOINT (str): defines the endpoint for which to apply this
            mapper.
        ALIASES (Tuple[Tuple[str, str]]): a tuple of aliases between
            OPTIMADE field names and the field names in the database ,
            e.g. `(("elements", "custom_elements_field"))`.
        LENGTH_ALIASES (Tuple[Tuple[str, str]]): a tuple of aliases between
            a field name and another field that defines its length, to be used
            when querying, e.g. `(("elements", "nelements"))`.
            e.g. `(("elements", "custom_elements_field"))`.
        PROVIDER_FIELDS (Tuple[str]): a tuple of extra field names that this
            mapper should support when querying with the database prefix.
        REQUIRED_FIELDS (set[str]): the set of fieldnames to return
            when mapping to the OPTIMADE format.
        TOP_LEVEL_NON_ATTRIBUTES_FIELDS (set[str]): the set of top-level
            field names common to all endpoints.

    """

    ENDPOINT: str = ""
    ALIASES: Tuple[Tuple[str, str]] = ()
    LENGTH_ALIASES: Tuple[Tuple[str, str]] = ()
    PROVIDER_FIELDS: Tuple[str] = ()
    REQUIRED_FIELDS: set = set()
    TOP_LEVEL_NON_ATTRIBUTES_FIELDS: set = {"id", "type", "relationships", "links"}

    @classmethod
    def all_aliases(cls) -> Tuple[Tuple[str, str]]:
        """ Returns all of the associated aliases for this class, including
        those defined by the server config.

        Returns:
            Tuple[Tuple[str, str]]: a tuple of alias tuples.

        """
        return (
            tuple(
                (f"_{CONFIG.provider.prefix}_{field}", field)
                for field in CONFIG.provider_fields.get(cls.ENDPOINT, [])
            )
            + tuple(
                (f"_{CONFIG.provider.prefix}_{field}", field)
                for field in cls.PROVIDER_FIELDS
            )
            + tuple(CONFIG.aliases.get(cls.ENDPOINT, {}).items())
            + cls.ALIASES
        )

    @classmethod
    def all_length_aliases(cls) -> Tuple[Tuple[str, str]]:
        """ Returns all of the associated length aliases for this class, including
        those defined by the server config.

        Returns:
            Tuple[Tuple[str, str]]: a tuple of length alias tuples.

        """
        return cls.LENGTH_ALIASES + tuple(
            CONFIG.length_aliases.get(cls.ENDPOINT, {}).items()
        )

    @classmethod
    def length_alias_for(cls, field: str) -> str:
        """ Returns the length alias for the particular field, or `None` if no
        such alias is found.

        """
        return dict(cls.all_length_aliases()).get(field, None)

    @classmethod
    def alias_for(cls, field: str) -> str:
        """Return aliased field name.

        :param field: OPTIMADE field name
        :type field: str

        :return: Aliased field as found in PROVIDER_ALIASES + ALIASES
        :rtype: str
        """
        split = field.split(".")
        alias = dict(cls.all_aliases()).get(split[0], None)
        if alias is not None:
            return alias + ("." + ".".join(split[1:]) if len(split) > 1 else "")
        return field

    @classmethod
    def get_required_fields(cls) -> set:
        """Return set REQUIRED response fields"""
        res = cls.TOP_LEVEL_NON_ATTRIBUTES_FIELDS.copy()
        res.update(cls.REQUIRED_FIELDS)
        return res

    @classmethod
    def map_back(cls, doc: dict) -> dict:
        """Map properties from MongoDB to OPTIMADE

        Starting from a MongoDB document ``doc``, map the DB fields to the corresponding OPTIMADE fields.
        Then, the fields are all added to the top-level field "attributes",
        with the exception of other top-level fields, defined in ``cls.TOPLEVEL_NON_ATTRIBUTES_FIELDS``.
        All fields not in ``cls.TOPLEVEL_NON_ATTRIBUTES_FIELDS`` + "attributes" will be removed.
        Finally, the ``type`` is given the value of the specified ``cls.ENDPOINT``.

        :param doc: A resource object in MongoDB format
        :type doc: dict

        :return: A resource object in OPTIMADE format
        :rtype: dict

        """
        if "_id" in doc:
            del doc["_id"]

        mapping = ((real, alias) for alias, real in cls.all_aliases())
        newdoc = {}
        reals = {real for alias, real in cls.all_aliases()}
        for key in doc:
            if key not in reals:
                newdoc[key] = doc[key]
        for real, alias in mapping:
            if real in doc:
                newdoc[alias] = doc[real]

        if "attributes" in newdoc:
            raise Exception("Will overwrite doc field!")
        attributes = newdoc.copy()

        for field in cls.TOP_LEVEL_NON_ATTRIBUTES_FIELDS:
            value = attributes.pop(field, None)
            if value is not None:
                newdoc[field] = value
        for field in list(newdoc.keys()):
            if field not in cls.TOP_LEVEL_NON_ATTRIBUTES_FIELDS:
                del newdoc[field]

        newdoc["type"] = cls.ENDPOINT
        newdoc["attributes"] = attributes

        return newdoc



# pylint: disable=no-self-argument
from pydantic import Field, AnyUrl, validator, root_validator
from typing import Union

from .jsonapi import Link, Attributes
from .entries import EntryResource


__all__ = (
    "LinksResourceAttributes",
    "LinksResource",
    "ChildResource",
    "ParentResource",
    "ProviderResource",
)


class LinksResourceAttributes(Attributes):
    """Links endpoint resource object attributes"""

    name: str = Field(
        ...,
        description="Human-readable name for the OPTIMADE API implementation "
        "a client may provide in a list to an end-user.",
    )
    description: str = Field(
        ...,
        description="Human-readable description for the OPTIMADE API implementation "
        "a client may provide in a list to an end-user.",
    )
    base_url: Union[AnyUrl, Link, None] = Field(
        ...,
        description="JSON API links object, pointing to the base URL for this implementation",
    )

    homepage: Union[AnyUrl, Link, None] = Field(
        ...,
        description="JSON API links object, pointing to a homepage URL for this implementation",
    )


class LinksResource(EntryResource):
    """A Links endpoint resource object"""

    type: str = Field(
        ...,
        description='MUST be either "parent", "child", or "provider". '
        "These objects are described in detail in sections Parent and Child Objects "
        "and Provider Objects.",
    )

    attributes: LinksResourceAttributes = Field(
        ...,
        description="a dictionary containing key-value pairs representing the "
        "entry's properties.",
    )

    @validator("type")
    def type_must_be_in_specific_set(cls, value):
        if value not in {"parent", "child", "provider"}:
            raise ValueError(
                "name of Links endpoint resource MUST be either 'parent, 'child', or 'provider'"
            )
        return value

    @root_validator(pre=True)
    def relationships_must_not_be_present(cls, values):
        if values.get("relationships", None) is not None:
            raise ValueError('"relationships" is not allowed for links resources')
        return values


class ChildResource(LinksResource):
    """A child object representing a link to an implementation exactly one layer below the current implementation"""

    type: str = Field("child", const=True)


class ParentResource(LinksResource):
    """A parent object representing a link to an implementation exactly one layer above the current implementation"""

    type: str = Field("parent", const=True)


class ProviderResource(LinksResource):
    """A provider object representing a link to another index meta-database by another database provider"""

    type: str = Field("provider", const=True)



# Copyright 2018 Markus Scheidgen, Alvin Noe Ladines
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an"AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Iterable, Any, Tuple, Dict, Callable, BinaryIO, Union, List, cast
from io import BytesIO, BufferedReader
from collections.abc import Mapping, Sequence
import msgpack
from msgpack.fallback import Packer, StringIO
import struct
import json
import math
import re

from nomad import utils, config, infrastructure
from nomad.metainfo import Quantity, Reference, Definition, MSection, Section, SubSection
from nomad.datamodel import EntryArchive
from nomad.datamodel.metainfo.public import fast_access

'''
The archive storage is made from two tiers. First the whole archive is stored in
files, secondly parts of the archive are stored in mongodb documents.

The file storage is done in msg-pack files. Each file contains the archive of many
entries (all entries of an upload). These msg-pack files contain the JSON serialized
version of the metainfo archive (see module:`nomad.metainfo`). In addition msg-pack
contains TOC information for quicker access of individual sections. See :func:`write_archive`
and :func:`read_archvive`. In addition there query functionality to partially read
specified sections from an archive: func:`query_archive`.

The mongo storage uses mongodb's native bson to store JSON serialized metainfo archive
data. Each document in mongodb holds the partial archive of single entry. Which parts
of an archive are stored in mongo is determined by the metainfo and
section annotations/categories.
'''

__packer = msgpack.Packer(autoreset=True, use_bin_type=True)


def packb(o, **kwargs):
    return __packer.pack(o)


def unpackb(o, **kwargs):
    return msgpack.unpackb(o, raw=False)


def adjust_uuid_size(uuid):
    uuid = uuid.rjust(utils.default_hash_len, ' ')
    assert len(uuid) == utils.default_hash_len, 'uuids must have the right fixed size'
    return uuid


class ArchiveError(Exception):
    ''' An error that indicates a broken archive. '''
    pass


class ArchiveQueryError(Exception):
    '''
    An error that indicates that an archive query is either not valid or does not fit to
    the queried archive.
    '''
    pass


class TOCPacker(Packer):
    '''
    A special msgpack packer that records a TOC while packing.

    Uses a combination of the pure python msgpack fallback packer and the "real"
    c-based packing.
    '''
    def __init__(self, toc_depth: int, *args, **kwargs):
        self.toc_depth = toc_depth
        self.toc: Dict[str, Any] = None
        self._depth = 0

        # Because we cannot change msgpacks interface of _pack, this _stack is used to
        # tranfer the result of _pack calls in terms of the TOC.
        self._stack: List[Any] = []

        super().__init__(*args, **kwargs)

    def pack(self, obj, *args, **kwargs):
        assert isinstance(obj, dict), 'TOC packer can only pack dicts, %s' % obj.__class__
        self._depth = 0
        self._buffer = StringIO()
        result = super().pack(obj, *args, **kwargs)
        self.toc = self._stack.pop()
        assert len(self._stack) == 0
        return result

    def _pos(self):
        return self._buffer.getbuffer().nbytes

    def _pack(self, obj, *args, **kwargs):
        if isinstance(obj, dict):
            toc_result = {}
            start = self._pos()
            if self._depth >= self.toc_depth:
                pack_result = self._buffer.write(packb(obj))
            else:
                self._depth += 1
                pack_result = super()._pack(obj, *args, **kwargs)
                self._depth -= 1

                toc = {}
                for key, value in reversed(list(obj.items())):
                    if isinstance(value, (list, tuple)):
                        # this makes some assumptions about the non emptyness and uniformness
                        # of array items
                        if len(value) > 0 and isinstance(value[0], dict):
                            toc[key] = self._stack.pop()

                    elif isinstance(value, (dict, list, tuple)):
                        toc[key] = self._stack.pop()

                toc_result['toc'] = {key: value for key, value in reversed(list(toc.items()))}

            end = self._pos()
            toc_result['pos'] = [start, end]

            self._stack.append(toc_result)

        elif isinstance(obj, list):
            toc_result = []
            pack_result = super()._pack(obj, *args, **kwargs)

            # same assumption and condition as above
            if len(obj) > 0 and isinstance(obj[0], dict):
                for _ in obj:
                    toc_result.append(self._stack.pop())

                self._stack.append(list(reversed(toc_result)))

        else:
            pack_result = self._buffer.write(packb(obj))

        return pack_result


_toc_uuid_size = utils.default_hash_len + 1
_toc_item_size = _toc_uuid_size + 25  # packed(uuid + [10-byte-pos, 10-byte-pos])


class ArchiveWriter:
    def __init__(self, file_or_path: Union[str, BytesIO], n_entries: int, entry_toc_depth: int):
        self.file_or_path = file_or_path
        self.n_entries = n_entries

        self._pos = 0
        self._toc_position: Tuple[int, int] = None
        self._toc: Dict[str, Tuple[Tuple[int, int], Tuple[int, int]]] = {}
        self._f: BinaryIO = None
        self._toc_packer = TOCPacker(toc_depth=entry_toc_depth)

    def __enter__(self):
        if isinstance(self.file_or_path, str):
            self._f = open(self.file_or_path, 'wb')
        elif isinstance(self.file_or_path, BytesIO):
            self._f = self.file_or_path
            self._f.seek(0)
        else:
            raise ValueError('not a file or path')

        # write empty placeholder header
        self._write_map_header(3)
        self.write(packb('toc_pos'))
        self.write(packb(ArchiveWriter._encode_position(0, 0)))

        self.write(packb('toc'))
        toc_start, _ = self._write_map_header(self.n_entries)
        _, toc_end = self.write(b'0' * _toc_item_size * self.n_entries)
        self._toc_position = toc_start, toc_end

        self.write(packb('data'))
        self._write_map_header(self.n_entries)

        return self

    def write(self, b: bytes) -> Tuple[int, int]:
        start = self._pos
        self._pos += self._f.write(b)
        return start, self._pos

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_val is not None:
            raise exc_val

        # go back and write the real TOC to the header
        self._f.seek(0)
        self._pos = 0

        assert len(self._toc) == self.n_entries
        toc_items = sorted(self._toc.items(), key=lambda item: item[0])
        toc = {
            uuid: [
                ArchiveWriter._encode_position(*positions[0]),
                ArchiveWriter._encode_position(*positions[1])]
            for uuid, positions in toc_items}

        self._write_map_header(3)
        self.write(packb('toc_pos'))
        self.write(
            packb(ArchiveWriter._encode_position(*self._toc_position), use_bin_type=True))

        self.write(packb('toc'))
        toc_position = self.write(packb(toc, use_bin_type=True))
        assert toc_position == self._toc_position, '%s - %s' % (toc_position, self._toc_position)

        if isinstance(self.file_or_path, str):
            self._f.close()

    @staticmethod
    def _encode_position(start: int, end: int) -> bytes:
        return start.to_bytes(5, byteorder='little', signed=False) + \
            end.to_bytes(5, byteorder='little', signed=False)

    def add(self, uuid: str, data: Any) -> None:
        uuid = adjust_uuid_size(uuid)

        self._toc_packer.reset()
        packed = self._toc_packer.pack(data)
        toc = self._toc_packer.toc

        self.write(packb(uuid))
        self._write_map_header(2)
        self.write(packb('toc'))
        toc_pos = self.write(packb(toc, use_bin_type=True))
        self.write(packb('data'))
        data_pos = self.write(packed)

        self._toc[uuid] = (toc_pos, data_pos)

    def _write_map_header(self, n):
        if n <= 0x0f:
            return self.write(struct.pack('B', 0x80 + n))
        if n <= 0xffff:
            return self.write(struct.pack(">BH", 0xde, n))
        if n <= 0xffffffff:
            return self.write(struct.pack(">BI", 0xdf, n))
        raise ValueError("Dict is too large")


class ArchiveItem:
    def __init__(self, toc_entry: list, f: BytesIO, offset: int = 0):
        self.toc_entry = toc_entry
        self._f = f
        self._offset = offset

    def _read(self, position: Tuple[int, int]):
        start, end = position
        self._f.seek(start + self._offset)
        return unpackb(self._f.read(end - start))

    def _getchild(self, child_toc_entry):
        if isinstance(child_toc_entry, dict):
            if 'toc' in child_toc_entry:
                return ArchiveObject(child_toc_entry, self._f, self._offset)
            else:
                return self._read(child_toc_entry['pos'])

        elif isinstance(child_toc_entry, list):
            return ArchiveList(child_toc_entry, self._f, self._offset)

        else:
            assert False, 'unreachable'


class ArchiveList(ArchiveItem, Sequence):

    def __getitem__(self, index):
        child_toc_entry = self.toc_entry[index]
        return self._getchild(child_toc_entry)

    def __len__(self):
        return self.toc_entry.__len__()


class ArchiveObject(ArchiveItem, Mapping):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._data = None

    def __getitem__(self, key):
        if self._data is None:
            try:
                child_toc_entry = self.toc_entry['toc'][key]
            except KeyError:
                return self.to_dict().__getitem__(key)
            return self._getchild(child_toc_entry)
        else:
            return self._data.__getitem__(key)

    def __iter__(self):
        if self._data is None:
            return self.toc_entry['toc'].__iter__()
        else:
            return self._data.__iter__()

    def __len__(self):
        if self._data is None:
            return self.toc_entry['toc'].__len__()
        else:
            return self._data.__len__()

    def to_dict(self):
        return self._read(self.toc_entry['pos'])


class ArchiveReader(ArchiveObject):
    def __init__(self, file_or_path: Union[str, BytesIO], use_blocked_toc=True):
        self.file_or_path = file_or_path

        f: BytesIO = None
        if isinstance(self.file_or_path, str):
            f = cast(BytesIO, open(self.file_or_path, 'rb'))
        elif isinstance(self.file_or_path, BytesIO):
            f = self.file_or_path
        elif isinstance(self.file_or_path, BufferedReader):
            f = self.file_or_path
        else:
            raise ValueError('not a file or path')

        super().__init__(None, f)

        self._toc_entry = None

        # this number is determined by the msgpack encoding of the file beginning:
        # { 'toc_pos': <...>
        #              ^11
        self._f.seek(11)
        self.toc_position = ArchiveReader._decode_position(self._f.read(10))

        self.use_blocked_toc = use_blocked_toc
        if use_blocked_toc:
            self._f.seek(11)
            self._toc: Dict[str, Any] = {}
            toc_start = self.toc_position[0]
            self._f.seek(toc_start)
            b = self._f.read(1)[0]
            if b & 0b11110000 == 0b10000000:
                self._n_toc = b & 0b00001111
                self._toc_offset = toc_start + 1
            elif b == 0xde:
                self._n_toc, = struct.unpack_from(">H", self._f.read(2))
                self._toc_offset = toc_start + 3
            elif b == 0xdf:
                self._n_toc, = struct.unpack_from(">I", self._f.read(4))
                self._toc_offset = toc_start + 5
            else:
                raise ArchiveError('Archive top-level TOC is not a msgpack map (dictionary).')

        else:
            self.toc_entry = self._read(self.toc_position)

    def __enter__(self):
        return self

    fs_block_size = 4096
    toc_block_size_entries = math.ceil(4096 / _toc_item_size)
    toc_block_size_bytes = math.ceil(4096 / 54) * _toc_item_size

    def _load_toc_block(self, i_entry: int):
        i_block = math.floor(i_entry / ArchiveReader.toc_block_size_entries)

        self._f.seek(i_block * ArchiveReader.toc_block_size_bytes + self._toc_offset)
        block_data = self._f.read(ArchiveReader.toc_block_size_bytes)
        first, last = None, None
        toc_block_size_entries = min(
            ArchiveReader.toc_block_size_entries,
            self._n_toc - i_block * ArchiveReader.toc_block_size_entries)

        for i in range(0, toc_block_size_entries):
            offset = i * _toc_item_size
            entry_uuid = unpackb(block_data[offset:offset + _toc_uuid_size])
            positions_encoded = unpackb(block_data[offset + _toc_uuid_size:offset + _toc_item_size])
            positions = (
                ArchiveReader._decode_position(positions_encoded[0]),
                ArchiveReader._decode_position(positions_encoded[1]))
            self._toc[entry_uuid] = positions

            if i == 0:
                first = entry_uuid

            if i + 1 == toc_block_size_entries:
                last = entry_uuid

        return first, last

    def __getitem__(self, key):
        key = adjust_uuid_size(key)

        if self.use_blocked_toc and self.toc_entry is None:
            if self._n_toc == 0:
                raise KeyError(key)

            positions = self._toc.get(key)
            # TODO use hash algorithm instead of binary search
            if positions is None:
                r_start = 0
                r_end = self._n_toc
                i_block = None
                while r_start <= r_end:
                    new_i_block = r_start + math.floor((r_end - r_start) / 2)
                    if i_block == new_i_block:
                        break
                    else:
                        i_block = new_i_block

                    first, last = self._load_toc_block(i_block)

                    if key < first:
                        r_end = i_block - 1
                    elif key > last:
                        r_start = i_block + 1
                    else:
                        break

                positions = self._toc.get(key)
                if positions is None:
                    raise KeyError(key)

            toc_position, data_position = positions

        else:
            positions_encoded = self.toc_entry[key]
            toc_position = ArchiveReader._decode_position(positions_encoded[0])
            data_position = ArchiveReader._decode_position(positions_encoded[1])

        toc = self._read(toc_position)
        return ArchiveObject(toc, self._f, data_position[0])

    def __iter__(self):
        if self.toc_entry is None:
            # is not necessarily read when using blocked toc
            self.toc_entry = self._read(self.toc_position)

        return self.toc_entry.__iter__()

    def __len__(self):
        if self.toc_entry is None:
            # is not necessarily read when using blocked toc
            self.toc_entry = self._read(self.toc_position)

        return self.toc_entry.__len__()

    def close(self):
        if isinstance(self.file_or_path, str):
            self._f.close()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @staticmethod
    def _decode_position(position: bytes) -> Tuple[int, int]:
        return int.from_bytes(position[0:5], byteorder='little', signed=False), \
            int.from_bytes(position[5:], byteorder='little', signed=False)

    def is_closed(self):
        return self._f.closed


def write_archive(
        path_or_file: Union[str, BytesIO], n_entries: int, data: Iterable[Tuple[str, Any]],
        entry_toc_depth: int = 2) -> None:
    '''
    Writes a msgpack-based archive file. The file contents will be a valid msgpack-object.
    The data will contain extra table-of-contents (TOC) objects that map some keys to
    positions in the file. Data can be partially read from these positions and deserialized
    with msgpack.

    The data in the archive file will have the following layout:

    .. code-block:: python

        {
            'toc_pos': b[start, end],
            'toc': {
                entry_uuid: [b[start, end], b[start, end]], ...
            },
            'data': {
                entry_uuid: {
                    'toc': {
                        key: {
                            'pos': [start, end],
                            'toc': ...
                        },
                        key: [
                            {
                                'pos': [start, end]
                                'toc': ...
                            }, ...
                        ],
                        ...
                    },
                    'data': ...
                }, ...
            }
        }


    The top-level TOC will map entry_uuids to positions. The key 'toc_pos' holds the
    position of the entry TOC, the second ('toc') the position of each entry. These positions
    will be absolute positions in the file. The top-level TOC will be ordered by entry_uuid.
    The top-level TOC positions are 2*5byte encoded integers. This will give the top-level TOC a
    predictable layout and will allow to partially read this TOC.

    The TOC of each entry will have the same structure than the data up to a certain
    TOC depth. A TOC object will hold the position of the object it refers to (key 'pos')
    and further deeper TOC data (key 'toc'). Only data objects (dict instances) will
    have TOC objects and only object count towards the TOC depth. Positions in the entry
    TOCs are regular msgpack encoded integers.

    Arguments:
        file_or_path: A file path or file-like to the archive file that should be written.
        n_entries: The number of entries that will be added to the file.
        data: The file contents as an iterator of entry id, data tuples.
        entry_toc_depth: The depth of the table of contents in each entry. Only objects will
            count for calculating the depth.
    '''
    with ArchiveWriter(path_or_file, n_entries, entry_toc_depth=entry_toc_depth) as writer:
        for uuid, entry in data:
            writer.add(uuid, entry)


def read_archive(file_or_path: str, **kwargs) -> ArchiveReader:
    '''
    Allows to read a msgpack-based archive.

    Arguments:
        file_or_path: A file path or file-like to the archive file that should be read. The
            respective file has to be closed by the user. The returned obj supports the
            'with' statement and has a 'close' method.

    Returns:
        A mapping (dict-like) that can be used to access the archive data. The mapping
        will lazily load data as it is used. The mapping needs to be closed or used within
        a 'with' statement to free the underlying file resource after use.
    '''

    return ArchiveReader(file_or_path, **kwargs)


__query_archive_key_pattern = re.compile(r'^([\s\w\-]+)(\[([-?0-9]*)(:([-?0-9]*))?\])?$')


def query_archive(
        f_or_archive_reader: Union[str, ArchiveReader, BytesIO], query_dict: dict,
        **kwargs) -> Dict:
    '''
    Takes an open msg-pack based archive (either as str, reader, or BytesIO) and returns
    the archive as JSON serializable dictionary filtered based on the given required
    specification.
    '''
    def _to_son(data):
        if isinstance(data, (ArchiveList, List)):
            data = [_to_son(item) for item in data]

        elif isinstance(data, ArchiveObject):
            data = data.to_dict()

        return data

    def _load_data(query_dict: Dict[str, Any], archive_item: ArchiveObject) -> Dict:
        query_dict_with_fixed_ids = {
            adjust_uuid_size(key): value for key, value in query_dict.items()}
        return filter_archive(query_dict_with_fixed_ids, archive_item, transform=_to_son)

    if isinstance(f_or_archive_reader, ArchiveReader):
        return _load_data(query_dict, f_or_archive_reader)

    elif isinstance(f_or_archive_reader, (BytesIO, str)):
        with ArchiveReader(f_or_archive_reader, **kwargs) as archive:
            return _load_data(query_dict, archive)

    else:
        raise TypeError('%s is neither a file-like nor ArchiveReader' % f_or_archive_reader)


def filter_archive(
        required: Dict[str, Any], archive_item: Union[Dict, ArchiveObject],
        transform: Callable) -> Dict:

    def _fix_index(index, length):
        if index is None:
            return index
        if index < 0:
            return max(-(length), index)
        else:
            return min(length, index)

    if archive_item is None:
        return None

    if not isinstance(required, dict):
        return transform(archive_item)

    result: Dict[str, Any] = {}
    for key, val in required.items():
        key = key.strip()

        # process array indices
        match = __query_archive_key_pattern.match(key)
        index: Union[Tuple[int, int], int] = None
        if match:
            key = match.group(1)

            # check if we have indices
            if match.group(2) is not None:
                first_index, last_index = None, None
                group = match.group(3)
                first_index = None if group == '' else int(group)

                if match.group(4) is not None:
                    group = match.group(5)
                    last_index = None if group == '' else int(group)
                    index = (0 if first_index is None else first_index, last_index)

                else:
                    index = first_index  # one item

            else:
                index = None

        else:
            raise ArchiveQueryError('invalid key format: %s' % key)

        try:
            archive_child = archive_item[key]
            is_list = isinstance(archive_child, (ArchiveList, list))

            if index is None and is_list:
                index = (0, None)

            elif index is not None and not is_list:
                raise ArchiveQueryError('cannot use list key on none list %s' % key)

            if index is None:
                pass
            else:
                length = len(archive_child)
                if isinstance(index, tuple):
                    index = (_fix_index(index[0], length), _fix_index(index[1], length))
                    if index[0] == index[1]:
                        archive_child = [archive_child[index[0]]]
                    else:
                        archive_child = archive_child[index[0]: index[1]]
                else:
                    archive_child = [archive_child[_fix_index(index, length)]]

            if isinstance(archive_child, (ArchiveList, list)):
                result[key] = [
                    filter_archive(val, item, transform=transform)
                    for item in archive_child]
            else:
                result[key] = filter_archive(val, archive_child, transform=transform)

        except (KeyError, IndexError):
            continue

    return result


def create_partial_archive(archive: EntryArchive) -> Dict:
    '''
    Creates a partial archive JSON serializable dict that can be stored directly.
    The given archive is filtered based on the metainfo category ``fast_access``.
    Selected sections and other data that they reference (recursively) comprise the
    resulting partial archive.

    TODO at the moment is hard coded and NOT informed by the metainfo. We simply
    add sections EntryMetadata and Workflow.

    Arguments:
        archive: The archive as an :class:`EntryArchive` instance.

    Returns: the partial archive in JSON serializable dictionary form.
    '''
    # A list with all referenced sections that might not yet been ensured to be in the
    # resulting partial archive
    referenceds: List[MSection] = []
    # contents keeps track of all sections in the partial archive by keeping their
    # JSON serializable form and placeholder status in a dict
    contents: Dict[MSection, Tuple[dict, bool]] = dict()

    def partial(definition: Definition, section: MSection) -> bool:
        '''
        ``m_to_dict`` partial function that selects what goes into the partial archive
        and what not. It also collects references as a side-effect.
        '''
        if section.m_def == EntryArchive.m_def:
            if definition.m_def == Quantity:
                return True
            return fast_access.m_def in definition.categories

        if isinstance(definition, Quantity) and isinstance(definition.type, Reference) \
                and fast_access.m_def in definition.categories:
            # Reference list in partial archives are not supported
            if definition.is_scalar:
                referenced = getattr(section, definition.name)
                if referenced is not None and referenced not in contents:
                    referenceds.append(referenced)

        if isinstance(definition, SubSection):
            return fast_access.m_def in definition.categories

        return True

    # add the main content
    partial_contents = archive.m_to_dict(partial=partial)

    # add the referenced data
    def add(section, placeholder=False) -> dict:
        '''
        Adds the given section to partial_contents at the right place. If not a placeholder,
        the section's serialization is added (or replacing an existing placeholder).
        Otherwise, an empty dict is added as a placeholder for referenced children.
        '''
        result: Dict[str, Any] = None
        content, content_is_placeholder = contents.get(section, (None, True))
        if content is not None:
            if content_is_placeholder and not placeholder:
                # the placeholder gets replaced later
                pass
            else:
                return content

        if section.m_parent is None:
            contents[section] = partial_contents, False
            return partial_contents

        parent_dict = add(section.m_parent, placeholder=True)
        if placeholder:
            result = {}
        else:
            result = section.m_to_dict(partial=partial)

        sub_section = section.m_parent_sub_section
        if sub_section.repeats:
            sections = parent_dict.setdefault(sub_section.name, [])
            while len(sections) < section.m_parent_index + 1:
                sections.append(None)
            sections[section.m_parent_index] = result
        else:
            parent_dict[sub_section.name] = result

        contents[section] = result, placeholder
        return result

    # we add referenced objects as long as they are added by subsequent serialization
    # of referenced sections to implement the recursive nature of further references in
    # already referenced sections.
    while len(referenceds) > 0:
        referenced = referenceds.pop()
        add(referenced)

    return partial_contents


def write_partial_archive_to_mongo(archive: EntryArchive):
    ''' Partially writes the given archive to mongodb. '''
    mongo_db = infrastructure.mongo_client[config.mongo.db_name]
    mongo_collection = mongo_db['archive']
    mongo_id = archive.section_metadata.calc_id

    partial_archive_dict = create_partial_archive(archive)
    partial_archive_dict['_id'] = mongo_id
    mongo_collection.replace_one(dict(_id=mongo_id), partial_archive_dict, upsert=True)


def read_partial_archive_from_mongo(entry_id: str, as_dict=False) -> Union[EntryArchive, Dict]:
    '''
    Reads the partial archive for the given id from mongodb.

    Arguments:
        entry_id: The entry id for the entry.
        as_dict: Return the JSON serializable dictionary form of the archive not the
            :class:`EntryArchive` form.
    '''
    mongo_db = infrastructure.mongo_client[config.mongo.db_name]
    mongo_collection = mongo_db['archive']
    archive_dict = mongo_collection.find_one(dict(_id=entry_id))

    if as_dict:
        return archive_dict

    return EntryArchive.m_from_dict(archive_dict)


def read_partial_archives_from_mongo(entry_ids: List[str], as_dict=False) -> Dict[str, Union[EntryArchive, Dict]]:
    '''
    Reads the partial archives for a set of entries of the same upload.

    Arguments:
        entry_ids: A list of entry ids.
        as_dict: Return the JSON serializable dictionary form of the archive not the
            :class:`EntryArchive` form.

    Returns:
        A dictionary with entry_ids as keys.
    '''
    mongo_db = infrastructure.mongo_client[config.mongo.db_name]
    mongo_collection = mongo_db['archive']
    archive_dicts = mongo_collection.find(dict(_id={'$in': entry_ids}))

    if as_dict:
        return {archive_dict.pop('_id'): archive_dict for archive_dict in archive_dicts}

    return {
        archive_dict.pop('_id'): EntryArchive.m_from_dict(archive_dict)
        for archive_dict in archive_dicts}


__all_parent_sections: Dict[Section, Tuple[str, Section]] = {}


def _all_parent_sections():
    if len(__all_parent_sections) == 0:
        def add(section):
            for sub_section in section.all_sub_sections.values():
                sub_section_section = sub_section.sub_section.m_resolved()
                __all_parent_sections.setdefault(sub_section_section, []).append((sub_section.qualified_name(), section, ))
                add(sub_section_section)

        add(EntryArchive.m_def)

    return __all_parent_sections


class _Incomplete(Exception): pass


def compute_required_with_referenced(required):
    '''
    Updates the given required dictionary to ensure that references to non required
    sections within a partial fast access archive are included. Only references that
    are directly contained in required are added. References from wildcarded sub sections
    are ignored.

    Returns: A new required dict or None. None is returned if it is unclear if the required
    is only accessing information of fast access partial archives.
    '''
    # TODO this function should be based on the metainfo

    if not isinstance(required, dict):
        return required

    if any(key.startswith('section_run') for key in required):
        return None

    required = dict(**required)

    def add_parent_section(section, child_required):
        parent_sections = _all_parent_sections().get(section, [])
        if len(parent_sections) == 0:
            return [required]

        result = []
        for name, parent_section in parent_sections:
            child_key = name.split('.')[-1]
            for parent_required in add_parent_section(parent_section, None):
                result.append(parent_required.setdefault(child_key, child_required if child_required else {}))

        return result

    def traverse(
            current: Union[dict, str],
            parent: Section = EntryArchive.m_def):

        if isinstance(current, str):
            return

        for key, value in current.items():
            prop = key.split('[')[0]
            prop_definition = parent.all_properties[prop]
            if isinstance(prop_definition, SubSection):
                if fast_access.m_def not in prop_definition.categories:
                    raise _Incomplete()

                traverse(value, prop_definition.sub_section)
            if isinstance(prop_definition, Quantity) and isinstance(prop_definition.type, Reference):
                current[prop] = '*'
                if fast_access.m_def not in prop_definition.categories:
                    continue

                target_section_def = prop_definition.type.target_section_def.m_resolved()
                add_parent_section(target_section_def, value)
                traverse(value, target_section_def)

    try:
        traverse(dict(**required))
    except _Incomplete:
        # We realized that something is required that is not in the partial archive
        return None

    return required


if __name__ == '__main__':

    def benchmark():
        from time import time
        import sys

        with open('archive_test.json') as f:
            example_data = json.load(f)

        size = 5000 if len(sys.argv) == 1 else int(sys.argv[1])
        access_every = 2
        example_archive = [(utils.create_uuid(), example_data) for _ in range(0, size)]
        example_uuid = example_archive[int(size / 2)][0]

        # this impl
        # create archive
        start = time()
        buffer = BytesIO()
        write_archive(buffer, len(example_archive), example_archive, entry_toc_depth=2)
        print('archive.py: create archive (1): ', time() - start)

        # read single entry from archive
        buffer = BytesIO(buffer.getbuffer())
        for use_blocked_toc in [False, True]:
            start = time()
            for _ in range(0, 23):
                read_archive(buffer, use_blocked_toc=use_blocked_toc)[example_uuid]['section_run']['section_system']
            print('archive.py: access single entry system (23), blocked %d: ' % use_blocked_toc, (time() - start) / 23)

        # read every n-ed entry from archive
        buffer = BytesIO(buffer.getbuffer())
        for use_blocked_toc in [False, True]:
            start = time()
            for _ in range(0, 23):
                with read_archive(buffer, use_blocked_toc=use_blocked_toc) as data:
                    for i, entry in enumerate(example_archive):
                        if i % access_every == 0:
                            data[entry[0]]['section_run']['section_system']
            print('archive.py: access every %d-ed entry single entry system (23), blocked %d: ' % (access_every, use_blocked_toc), (time() - start) / 23)

        # just msgpack
        start = time()
        packb(example_archive)
        print('msgpack: create archive (1): ', time() - start)

    benchmark()



# Copyright 2015-2018 Lauri Himanen, Fawzi Mohamed, Ankit Kariryaa
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

""" Main parser file to parse simulation files created using CP2K DFT program.

This file creates the parser class for CP2K simulation files. The class extends
the Parser Interface class created for NOMAD-fair. This class also uses structlog
to increase analytics on logged errors during parsing uploading data to NOMAD.
"""

from builtins import next
from builtins import range
import os
import re
import logging
import importlib
from nomadcore.baseclasses import ParserInterface

# Needs to be imported in order for the importlib calls to work in python 2.7
import cp2kparser.versions.cp2k262.singlepointparser

logger = logging.getLogger("nomad")


class CP2KParser(ParserInterface):
    """This class handles the initial setup before any parsing can happen. It
    determines which version of CP2K was used to generate the output and then
    sets up a correct main parser.

    After the implementation has been setup, you can parse files with
    parse().
    """
    def __init__(self, metainfo_to_keep=None, backend=None, default_units=None, metainfo_units=None, debug=False, log_level=logging.ERROR, store=True):
        super(CP2KParser, self).__init__(metainfo_to_keep, backend, default_units, metainfo_units, debug, log_level, store)

    def setup_version(self):
        """Setups the version by looking at the output file and the version
        specified in it.
        """
        # Search for the CP2K version specification and the RUN_TYPE for the
        # calculation. The correct and optimized parser is initialized based on
        # this information.
        regex_version = re.compile(r" CP2K\| version string:\s+CP2K version ([\d\.]+)")
        regex_run_type = re.compile(r"\s+GLOBAL\| Run type\s+(.+)")
        n_lines = 100
        version_id = None
        run_type = None
        with open(self.parser_context.main_file, 'r') as outputfile:
            for i_line in range(n_lines):
                try:
                    line = next(outputfile)
                except StopIteration:
                    break
                result_version = regex_version.match(line)
                result_run_type = regex_run_type.match(line)
                if result_version:
                    version_id = result_version.group(1).replace('.', '')
                if result_run_type:
                    run_type = result_run_type.group(1)

        if version_id is None:
            msg = "Could not find a version specification from the given main file."
            logger.exception(msg)
            raise RuntimeError(msg)
        if run_type is None:
            msg = "Could not find a version specification from the given main file."
            logger.exception(msg)
            raise RuntimeError(msg)

        # Setup the root folder to the fileservice that is used to access files
        dirpath, filename = os.path.split(self.parser_context.main_file)
        dirpath = os.path.abspath(dirpath)
        self.parser_context.file_service.setup_root_folder(dirpath)
        self.parser_context.file_service.set_file_id(filename, "output")

        # Setup the correct main parser based on the version id. If no match
        # for the version is found, use the main parser for CP2K 2.6.2
        self.setup_main_parser({"version_id": version_id, "run_type": run_type})

    @staticmethod
    def get_mainfile_regex():
        regex_str = (
            "  \*\*\*\* \*\*\*\* \*\*\*\*\*\*  \*\*  PROGRAM STARTED AT\s.*\n"
            " \*\*\*\*\* \*\* \*\*\*  \*\*\* \*\*   PROGRAM STARTED ON\s*.*\n"
            " \*\*    \*\*\*\*   \*\*\*\*\*\*    PROGRAM STARTED BY .*\n"
            " \*\*\*\*\* \*\*    \*\* \*\* \*\*   PROGRAM PROCESS ID .*\n"
            "  \*\*\*\* \*\*  \*\*\*\*\*\*\*  \*\*  PROGRAM STARTED IN .*\n"
        )
        return regex_str

    def get_metainfo_filename(self):
        return "cp2k.nomadmetainfo.json"

    def get_parser_info(self):
        return {'name': 'cp2k-parser', 'version': '1.0'}

    def setup_main_parser(self, version_dictionary):
        """
        Setups a main parser class for this calculation. The main class can be
        different for each version and run type.

        Args:
            version_id: An integer representing the CP2K version. The version
                number is originally a string the form '2.6.2', but here the
                numbers are just concatenated into a single integer number 262.
            run_type: A string that identifies the RUN_TYPE for the
                calculation.  All the possible run types can be found in the
                CP2K reference manual.

        Returns:
            A python class that should be instantiated later with the correct
            parameters.
        """
        run_type = version_dictionary["run_type"]
        version_id = version_dictionary["version_id"]

        # Search for a RUN_TYPE specific parser
        parser_map = {
            "ENERGY": "SinglePointParser",
            "ENERGY_FORCE": "SinglePointParser",
            "WAVEFUNCTION_OPTIMIZATION": "SinglePointParser",
            "WFN_OPT": "SinglePointParser",
            "GEO_OPT": "GeoOptParser",
            "GEOMETRY_OPTIMIZATION": "GeoOptParser",
            "MD": "MDParser",
            "MOLECULAR_DYNAMICS": "MDParser",
        }
        try:
            parser = parser_map[run_type]
        except KeyError:
            logger.exception(
                "A parser corresponding to the run_type '{}' could not be found."
                .format(run_type)
            )
            raise

        # Currently the version id is a pure integer, so it can directly be mapped
        # into a package name.
        base = "cp2kparser.versions.cp2k{}.{}".format(version_id, parser.lower())
        parser_module = None
        parser_class = None

        try:
            parser_module = importlib.import_module(base)
        except ImportError:
            logger.warning(
                "Could not find a parser for version '{}' and run type '{}'. "
                "Trying to default to the base implementation for CP2K 2.6.2"
                .format(version_id, run_type)
            )
            base = "cp2kparser.versions.cp2k262.{}".format(parser.lower())
            try:
                parser_module = importlib.import_module(base)
            except ImportError:
                logger.exception(
                    "Tried to default to the CP2K 2.6.2 implementation but "
                    "could not find the correct modules for run_type '{}'."
                    .format(run_type)
                )
                raise
        try:
            parser_class = getattr(parser_module, "CP2K{}".format(parser))
        except AttributeError:
            logger.exception(
                "A parser class '{}' could not be found in the module '[]'."
                .format(parser_class, parser_module)
            )
            raise

        self.main_parser = parser_class(self.parser_context)


from unittest import TestCase, skipIf

from optimade.filterparser import LarkParser

try:
    from optimade.filtertransformers.elasticsearch import ElasticTransformer, Quantity

    ELASTICSEARCH_IMPORTED = True
except ImportError:
    ELASTICSEARCH_IMPORTED = False


@skipIf(not ELASTICSEARCH_IMPORTED, "No ElasticSearch installation, skipping tests...")
class TestTransformer(TestCase):
    def setUp(self):
        self.parser = LarkParser(version=(0, 10, 1))

        nelements = Quantity(name="nelements")
        elements_only = Quantity(name="elements_only")
        elements_ratios = Quantity(name="elements_ratios")
        elements_ratios.nested_quantity = elements_ratios
        elements = Quantity(
            name="elements",
            length_quantity=nelements,
            has_only_quantity=elements_only,
            nested_quantity=elements_ratios,
        )
        dimension_types = Quantity(name="dimension_types")
        dimension_types.length_quantity = dimension_types

        quantities = [
            nelements,
            elements_only,
            elements_ratios,
            elements,
            dimension_types,
            Quantity(name="chemical_formula_reduced"),
        ]

        self.transformer = ElasticTransformer(quantities=quantities)

    def test_parse_n_transform(self):
        queries = [
            ("nelements > 1", 4),
            ("nelements >= 2", 4),
            ("nelements > 2", 1),
            ("nelements < 4", 4),
            ("nelements < 3", 3),
            ("nelements <= 3", 4),
            ("nelements != 2", 1),
            ("1 < nelements", 4),
            ('elements HAS "H"', 4),
            ('elements HAS ALL "H", "O"', 4),
            ('elements HAS ALL "H", "C"', 1),
            ('elements HAS ANY "H", "C"', 4),
            ('elements HAS ANY "C"', 1),
            ('elements HAS ONLY "C"', 0),
            ('elements HAS ONLY "H", "O"', 3),
            ('elements:elements_ratios HAS "H":>0.66', 2),
            ('elements:elements_ratios HAS ALL "O":>0.33', 3),
            ('elements:elements_ratios HAS ALL "O":>0.33,"O":<0.34', 2),
            ("elements IS KNOWN", 4),
            ("elements IS UNKNOWN", 0),
            ('chemical_formula_reduced = "H2O"', 2),
            ('chemical_formula_reduced CONTAINS "H2"', 3),
            ('chemical_formula_reduced CONTAINS "H"', 4),
            ('chemical_formula_reduced CONTAINS "C"', 1),
            ('chemical_formula_reduced STARTS "H2"', 3),
            ('chemical_formula_reduced STARTS WITH "H2"', 3),
            ('chemical_formula_reduced ENDS WITH "C"', 1),
            ('chemical_formula_reduced ENDS "C"', 1),
            ("elements LENGTH = 2", 3),
            ("elements LENGTH = 3", 1),
            ("dimension_types LENGTH = 0", 3),
            ("dimension_typesLENGTH = 1", 1),
            ("nelements = 2 AND dimension_types LENGTH = 1", 1),
            ("nelements = 3 AND dimension_types LENGTH = 1", 0),
            ("nelements = 3 OR dimension_types LENGTH = 1", 2),
            ("nelements > 1 OR dimension_types LENGTH = 1 AND nelements = 2", 4),
            ("(nelements > 1 OR dimension_types LENGTH = 1) AND nelements = 2", 3),
            ("NOT dimension_types LENGTH = 1", 3),
        ]

        for query, _ in queries:
            tree = self.parser.parse(query)
            result = self.transformer.transform(tree)
            self.assertIsNotNone(result)


if __name__ == "__main__":
    import unittest
    import sys

    suite = unittest.TestLoader().loadTestsFromModule(sys.modules[__name__])
    unittest.TextTestRunner().run(suite)
